#include <math.h>
#include <dos.h>
#include <fcntl.h>
#include <io.h>
#include <sys\types.h>
#include <sys\stat.h>
#include "pragmas.h"
#include "ves2.h"

//KENVEX.C:Last touch before disabling setdacbits(8) on 09/10/2006->09/11/2006 was 01/08/1999 3:50pm

#define SCISSORDIST 1.0

#define MAXXDIM 1600
#define MAXYDIM 1200

#define MAXREGIONS 1024
#define MAXPLANES 8192
#define MAXLOOPS 8192
#define MAXEDGES 16384
#define MAXVERTICES 8192

#define LOGBOXSIZ 3
#define BOXSIZ (1<<LOGBOXSIZ)
#define TILESIZ 64

#define MAXTILES 6144
#define MAXPALOOKUPS 64

#define BUILDMAXSECTORS 1024
#define BUILDMAXWALLS 8192
#define BUILDMAXSPRITES 4096

#define TIMERRATE (9942>>1)   //240 times/sec

#define PI 3.141592653589793
#define PI2 (PI*2)

typedef union { float f; long l; } kfloat;
typedef struct { float x, y, z; } point3d;

typedef struct
{
	long v;            //index to vertex
	long n;            //index to next edge (circular)
	long l;            //index to last edge (circular) (for speed only)
	long o;            //index to opposite edge
	long loop;         //index to loop of edge (for speed only)
} edgetype;

typedef struct
{
	long edgehead;     //index to primary edge on loop
	long n;            //index to next loop (circular)
	long plane;        //index to plane of loop (for speed only)
} looptype;

typedef struct
{
	long loophead;     //index to any loop in plane
	long n;            //index to next plane (circular)
	long o;            //index to opposite plane
	long region;       //index to region (for speed only)
	point3d normal;    //normal vector of plane (for speed only)
	short picnum;
	short alignmethod; //0:rel.align, 1:z-normal, 2:y-normal, 3:x-normal
	short uscale, vscale;
	unsigned short uoffs, voffs;
} planetype;

typedef struct
{
	long planehead;    //index to any plane in region
	char sortmethod, sortvalid, filler1, filler2;
	long tag;
} regiontype;

long getvertex(float, float, float);
long insideregion (point3d *, long);
long hitscan (long, point3d *, point3d *, point3d *);

	//Position variables
float ang1, ang2, ang3;
point3d ipos, ifor, istr, ihei;
long curregion = 0;

	//Board map variables
char boardfilename[80] = "boards.map";
long numregions, numplanes, numloops, numedges, numvertices;
regiontype region[MAXREGIONS];
planetype plane[MAXPLANES];
looptype loop[MAXLOOPS];
edgetype edge[MAXEDGES];
point3d vertex[MAXVERTICES];
char gotvertex[MAXVERTICES];

long planelist[MAXPLANES];
long regionlist[MAXREGIONS];
char gotregion[MAXREGIONS>>3];

static char pow2char[8] = {0x01,0x02,0x04,0x08,0x10,0x20,0x40,0x80};

long numsplitlines = 0, splitplane = -1, splitmode, joinplane = -1;
point3d splitline[MAXVERTICES];

	//Rotation/projection/rasterization variables
#define MAXPOINTS 1024
#define MAXBUNCHES 256
point3d rp[MAXVERTICES];
float sx[MAXVERTICES], sy[MAXVERTICES];
long point2[MAXVERTICES], point22[MAXVERTICES], split[MAXVERTICES];
char got[MAXPOINTS];
long lastx[MAXYDIM], nextx[MAXYDIM];

long bunchindex[MAXBUNCHES], numbunches;
typedef struct { long v, l; } bunchtype;
bunchtype bunch[MAXBUNCHES];
long sortlist[MAXBUNCHES], sortlistnum;

	//Span buffer variables
static long spbuf[131072], spcnt;
static long spi0, spi1, spi2;
static long spo0, spo1, spo2;

static long gtagcnt;

	//Screen variables
long xdim, ydim, hxdim, hydim;
float fxdim, fydim, fhxdim, fhydim, fxyaspect, fyxaspect, fhxdimyxaspect;
long ylookup[MAXYDIM+1];
long activepage = 0, visualpage = 0;
char overtbits;

	//Texture variables
long gtexturemode = 1, globpicnum = -1, oneregionmode = 0;
//char pic[TILESIZ*TILESIZ];
char actualdacbits;
long temppicnum = -1;

	//Palette variables
static char paletteloaded = 0, translucloaded = 0;
static short numpalookups;
static char palette[768], whitecol, blackcol;
static char palookup[MAXPALOOKUPS][256], *transluc;

	//Tiles???.ART variables
static char tempchar[16384];
static char *pic;
static long picsiz[MAXTILES];
static long waloff[MAXTILES], picanm[MAXTILES];
static long tilesizx[MAXTILES], tilesizy[MAXTILES];

typedef struct
{
	short wallptr, wallnum;
	long ceilingz, floorz;
	short ceilingstat, floorstat;
	short ceilingpicnum, ceilingheinum;
	signed char ceilingshade;
	char ceilingpal, ceilingxpanning, ceilingypanning;
	short floorpicnum, floorheinum;
	signed char floorshade;
	char floorpal, floorxpanning, floorypanning;
	char visibility, filler;
	short lotag, hitag, extra;
} buildsectortype;

typedef struct
{
	long x, y;
	short point2, nextwall, nextsector, cstat;
	short picnum, overpicnum;
	signed char shade;
	char pal, xrepeat, yrepeat, xpanning, ypanning;
	short lotag, hitag, extra;
} buildwalltype;

typedef struct
{
	long x, y, z;
	short cstat, picnum;
	signed char shade;
	char pal, clipdist, filler;
	unsigned char xrepeat, yrepeat;
	signed char xoffset, yoffset;
	short sectnum, statnum;
	short ang, owner, xvel, yvel, zvel;
	short lotag, hitag, extra;
} buildspritetype;

short buildnumsectors, buildnumwalls, buildnumsprites;
static buildsectortype buildsector[BUILDMAXSECTORS];
static buildwalltype buildwall[BUILDMAXWALLS];
//static buildspritetype buildsprite[BUILDMAXSPRITES];

long randomseed;

long searchplane = -1;
long searchx, searchy;
point3d searchpos;
long gridlock = 1, showgrid = 0;

	//FPS variables
long ofinetotalclock, ototalclock, chainintrclock = 0, averagefps;
#define AVERAGEFRAMES 32
long frameval[AVERAGEFRAMES], framecnt = 0;

#pragma aux inittimer42 =\
	"in al, 0x61",\
	"or al, 1",\
	"out 0x61, al",\
	"mov al, 0xb4",\
	"out 0x43, al",\
	"xor al, al",\
	"out 0x42, al",\
	"out 0x42, al",\
	modify exact [eax]\

#pragma aux uninittimer42 =\
	"in al, 0x61",\
	"and al, 252",\
	"out 0x61, al",\
	modify exact [eax]\

#pragma aux gettimer42 =\
	"mov al, 0x84",\
	"out 0x43, al",\
	"in al, 0x42",\
	"shl eax, 24",\
	"in al, 0x42",\
	"rol eax, 8",\
	modify [eax]\

long smallfnt[256] =
{
	0x000000,0x6f9f60,0x69f960,0xaffe40,0x4efe40,0x6ff6f0,0x66f6f0,0x000000,
	0xeeaee0,0x000000,0x000000,0x000000,0x000000,0x000000,0x7755c0,0x96f690,
	0x8cec80,0x26e620,0x4e4e40,0xaaa0a0,0x7dd550,0x7ca6c0,0x000ee0,0x4e4ee0,
	0x4e4440,0x444e40,0x02f200,0x04f400,0x000000,0x000000,0x000000,0x000000,
	0x000000,0x444040,0xaa0000,0xafafa0,0x6c46c0,0xa248a0,0x4a4ce0,0x240000,
	0x488840,0x422240,0x0a4a00,0x04e400,0x000224,0x00e000,0x000040,0x224480,
	0xeaaae0,0x444440,0xe2e8e0,0xe2e2e0,0xaae220,0xe8e2e0,0xe8eae0,0xe22220,
	0xeaeae0,0xeae220,0x040400,0x040480,0x248420,0x0e0e00,0x842480,0xc24040,
	0xeaece0,0x4aeaa0,0xcacac0,0x688860,0xcaaac0,0xe8c8e0,0xe8c880,0xe8aae0,
	0xaaeaa0,0xe444e0,0xe22a60,0xaacaa0,0x8888e0,0xaeeaa0,0xaeeea0,0xeaaae0,
	0xeae880,0xeaae60,0xeacaa0,0xe8e2e0,0xe44440,0xaaaae0,0xaaa440,0xaaeea0,
	0xaa4aa0,0xaae440,0xe248e0,0xc888c0,0x844220,0x622260,0x4a0000,0x0000e0,
	0x420000,0x006a60,0x88eae0,0x00e8e0,0x22eae0,0x006e60,0x24e440,0x06a62c,
	0x88eaa0,0x040440,0x040448,0x88aca0,0x444440,0x08eee0,0x00caa0,0x00eae0,
	0x00eae8,0x00eae2,0x00e880,0x0064c0,0x04e440,0x00aa60,0x00aa40,0x00eee0,
	0x00a4a0,0x00aa6c,0x00c460,0x648460,0x440440,0xc424c0,0x6c0000,0x04ae00,
	0x68886c,0xa0aa60,0x606e60,0xe06a60,0xa06a60,0xc06a60,0x046a60,0x00e8e4,
	0xe06e60,0xa06e60,0xc06e60,0x0a0440,0x0e0440,0x0c0440,0xa4aea0,0x404ea0,
	0x60ece0,0x007a70,0x7afab0,0xe0eae0,0xa0eae0,0xc0eae0,0xe0aa60,0xc0aa60,
	0xa0aa6c,0xa0eae0,0xa0aae0,0x4e8e40,0x65c4f0,0xa4ee40,0xcafab0,0x64e4c0,
	0x606a60,0x060440,0x60eae0,0x60aa60,0xc0caa0,0xe0aea0,0x6a60e0,0xeae0e0,
	0x404860,0x007400,0x00c400,0x8a4e60,0x8a6e20,0x404440,0x05a500,0x0a5a00,
	0x282828,0x5a5a5a,0xd7d7d7,0x444444,0x44c444,0x44cc44,0x66e666,0x00e666,
	0x00cc44,0x66ee66,0x666666,0x00ee66,0x66ee00,0x66e000,0x44cc00,0x00c444,
	0x447000,0x44f000,0x00f444,0x447444,0x00f000,0x44f444,0x447744,0x667666,
	0x667700,0x007766,0x66ff00,0x00ff66,0x667766,0x00ff00,0x66ff66,0x44ff00,
	0x66f000,0x00ff44,0x00f666,0x667000,0x447700,0x007744,0x007666,0x66f666,
	0x44ff44,0x44c000,0x007444,0xffffff,0x000fff,0xcccccc,0x333333,0xfff000,
	0x00dad0,0xcacac8,0xea8880,0x00f660,0xe848e0,0x007a40,0x0aac80,0x05a220,
	0xe4a4e0,0x4aea40,0x6996f0,0x646ae0,0x06f600,0x16f680,0x68c860,0x4aaaa0,
	0xe0e0e0,0x4e40e0,0x4240e0,0x4840e0,0x4a8880,0x222a40,0x40e040,0x06ec00,
	0xeae000,0x0cc000,0x00c000,0x644c40,0xcaa000,0xc4e000,0x0eee00,0x000000,
};

	//Timer handler variables
volatile long totalclock, lockclock, synctics;
void (__interrupt __far *oldtimerhandler)();
void __interrupt __far timerhandler(void);

	//Keyboard handler variables
volatile char keystatus[256], readch, oldreadch, extended, keytemp;
static void (__interrupt __far *oldkeyhandler)();
static void __interrupt __far keyhandler(void);

static long incx0, incx1, incy, incyy, globcol;
long uubuf[1024], uibuf[512], ubuf[512];
float gdo, guo, gvo, gdx, gux, gvx, gdy, guy, gvy;
float fgdxx, fguxx, fgvxx;


extern long prommxoverlays();
#pragma aux prommxoverlays modify [eax ebx ecx edx];
extern long setupkspin(long xybits, long bufplc);
#pragma aux setupkspin parm [eax][ebx] modify [ecx];
extern long kspin(long cnt, long, long incptr, long vidplc, long u, long v);
#pragma aux kspin parm [eax][ebx][ecx][edx][esi][edi];
extern long kspinshort(long cnt, long uinc, long vinc, long vidplc, long u, long v);
#pragma aux kspinshort parm [eax][ebx][ecx][edx][esi][edi];
extern long doratio(long ratio, long xind, long cnt);
#pragma aux doratio parm [eax][esi][edi] modify [ebx ecx edx];
extern long doratioinc(long xind, long cnt);
#pragma aux doratioinc parm [esi][edi] modify [eax ebx ecx edx];
extern long getubufasm1(long, long, long, long, long, long);
#pragma aux getubufasm1 parm [eax][ebx][ecx][edx][esi][edi];
extern long getubufasm2(long, long, long, long, long, long);
#pragma aux getubufasm2 parm [eax][ebx][ecx][edx][esi][edi];

#pragma aux msqrtasm =\
	"mov eax, 0x40000000",\
	"mov ebx, 0x20000000",\
	"begit: cmp ecx, eax",\
	"jl skip",\
	"sub ecx, eax",\
	"lea eax, [eax+ebx*4]",\
	"skip: sub eax, ebx",\
	"shr eax, 1",\
	"shr ebx, 2",\
	"jnz begit",\
	"cmp ecx, eax",\
	"sbb eax, -1",\
	"shr eax, 1",\
	parm nomemory [ecx]\
	modify exact [eax ebx ecx]\

	//0x007ff000 is (11<<13), 0x3f800000 is (127<<23)
long reciptable[2048], fpuasm;
#pragma aux krecipasm =\
	"mov fpuasm, eax",\
	"fild dword ptr fpuasm",\
	"add eax, eax",\
	"fstp dword ptr fpuasm",\
	"sbb ebx, ebx",\
	"mov eax, fpuasm",\
	"mov ecx, eax",\
	"and eax, 0x007ff000",\
	"shr eax, 10",\
	"sub ecx, 0x3f800000",\
	"shr ecx, 23",\
	"mov eax, dword ptr reciptable[eax]",\
	"sar eax, cl",\
	"xor eax, ebx",\
	parm [eax]\
	modify exact [eax ebx ecx]\

	//0x007fe000 is (10<<13), 0x44800000 is ((127+10)<<23)
#pragma aux krecipabsasm =\
	"fild dword ptr [eax]",\
	"fstp dword ptr fpuasm",\
	"mov eax, fpuasm",\
	"mov ecx, eax",\
	"and eax, 0x007ff000",\
	"shr eax, 10",\
	"sub ecx, 0x3f800000",\
	"shr ecx, 23",\
	"mov eax, dword ptr reciptable[eax]",\
	"sar eax, cl",\
	parm [eax]\
	modify exact [eax ecx]\

#pragma aux wcopybuf =\
	"rep movsw",\
	parm [esi][edi][ecx]\
	modify exact [ecx esi edi]\

#pragma aux wcopybufrev =\
	"std",\
	"rep movsw",\
	"cld",\
	parm [esi][edi][ecx]\
	modify exact [ecx esi edi]\

//Ŀ
// Rounding: D11-10:   Precision: D9-8:   
//Ĵ
// 00 = near/even (0)  00 = 24-bit    (0) 
// 01 = -inf      (4)  01 = reserved  (1) 
// 10 = +inf      (8)  10 = 53-bit    (2) 
// 11 = 0         (c)  11 = 64-bit    (3) 
//

long asm;
#pragma aux fpuinit =\
	"fninit",\
	"fstcw asm",\
	"mov ah, byte ptr asm[1]",\
	"and ah, 0f0h",\
	"or ah, al",\
	"mov byte ptr asm[1], ah",\
	"fldcw asm",\
	parm [eax]\

void ftol(double, long *);
#pragma aux ftol =\
	"fistp dword ptr [eax]",\
	parm [8087][eax]\

void fcossin(float, float *, float *);
#pragma aux fcossin =\
	"fsincos",\
	"fstp dword ptr [eax]",\
	"fstp dword ptr [ebx]",\
	parm [8087][eax][ebx]\

#pragma aux fdmultol =\
	"fld dword ptr [eax]",\
	"fmul st, st(1)",\
	"fxch st(1)",\
	"fld dword ptr [ebx]",\
	"fmulp st(1), st",\
	"fxch st(1)",\
	"fistp dword ptr [ecx]",\
	"fistp dword ptr [ecx+4]",\
	parm [eax][ebx][ecx]\

#pragma aux fpustartdiv =\
	"fld dword ptr [eax]",\
	"fdiv dword ptr [ebx]",\
	parm [eax][ebx]\
	modify exact \

#pragma aux fpugetnum =\
	"fstp dword ptr [eax]",\
	parm [eax]\
	modify exact \

#pragma aux fpusucknum =\
	"fstp st",\
	parm \
	modify exact \

main(int argc, char **argv)
{
	float ang1vel = 0, ang2vel = 0, ang3vel = 0, ox, oy, oz, f, r;
	float x0, y0, z0, x1, y1, z1, x2, y2, z2;
	long i, j, k, m, x, y, ix, iy;
	long lastreadmouse, p, startplane, l, ll, startloop, e, ee, startedge, dir;
	char tempbuf[256];
	short mousx, mousy, bstatus = 0, obstatus = 0;
	planetype *pla;
	point3d p0, p1, p2;

	prommxoverlays();

	if ((argc == 2) || (argc > 3))
		strcpy(boardfilename,argv[1]);

	if (argc >= 3)
		{ xdim = atoi(argv[argc-2]); ydim = atoi(argv[argc-1]); }
	else
	{
		getvalidvesamodes();
		for(i=validmodecnt-1;i>=0;i--)
			if ((validmodexdim[i] == 640) && (validmodeydim[i] == 480) && (validmodebits[i] == 8))
				{ xdim = 640; ydim = 480; break; }
		if (i < 0)
			if (validmodecnt > 0) { xdim = validmodexdim[0]; ydim = validmodeydim[0]; }
		else { xdim = 0; ydim = 0; }
	}
	if (setvesa(xdim,ydim,8) < 0)
	{
		if (argc >= 3)
		{
			printf("VESA says that %ld*%ld is not currently supported.\n",xdim,ydim);
			printf("Valid 256 color modes are:\n");
			getvalidvesamodes();
			for(i=0;i<validmodecnt;i++)
				if (validmodebits[i] == 8)
					printf("   %xh: %ld x %ld x %ld\n",validmode[i],validmodexdim[i],validmodeydim[i],validmodebits[i]);
			exit(0);
		}
		setvmode(0x13);
		xdim = 320; ydim = 200;
		frameplace = 0xa0000;
		bytesperline = 320;
		imageSize = 64000;
		buffermode = 1;
		maxpages = 1;
		if ((screen = (char *)malloc(imageSize)) == NULL)
		{
			printf("Not enough memory for VESA screen-buffering\n");
			exit(0);
		}
		frameplace = FP_OFF(screen);
		activepagelookup[0] = frameplace;
	}
	//if (maxpages > 3) maxpages = 3;

	actualdacbits = 6; //setdacbits(8);

	fpuinit(0x08);  //24 bit, round to +inf

	fxdim = (float)xdim; hxdim = (xdim>>1); fhxdim = ((float)xdim)*0.5;
	fydim = (float)ydim; hydim = (ydim>>1); fhydim = ((float)ydim)*0.5;
	fyxaspect = 1; //(fhydim/fhxdim)/(200.0/320.0);
	fxyaspect = 1 / fyxaspect;
	fhxdimyxaspect = fhxdim*fyxaspect;
	ylookup[0] = 0; for(i=1;i<=ydim;i++) ylookup[i] = ylookup[i-1]+bytesperline;

	searchx = hxdim; searchy = hydim;

	//setupmouse();

	for(i=0;i<2048;i++) reciptable[i] = divscale30(2048L,i+2048);

	if (vgacompatible)
	{
			//Init VR flag
		koutp(0x3d4,0x11);
		overtbits = kinp(0x3d5) & ~(16+32);
		koutp(0x3d5,overtbits);
		koutp(0x3d5,overtbits+16);
	}

	inittimer();
	inittimer42();
	initkeys();

	loadpics("tiles000.art");
	loadpalette();

	loadboard(boardfilename);

	lockclock = totalclock = lastreadmouse = 0;
	while (!keystatus[1])
	{
		synctics = totalclock-lockclock;
		lockclock += synctics;

		setactivepage(activepage);

		updateregion(&ipos,&curregion);

		drawscreen();

		for(i=(gtagcnt-1)*3;i>=0;i-=3)
		{
			j = spbuf[i+2]; if (i == j) continue;
			k = (i&63)+32; k += (k<<8); k += (k<<16);
			do
			{
				clearbufbyte(spbuf[j]+frameplace,spbuf[j+1]-spbuf[j],k);
				j = spbuf[j+2];
			} while (i != j);
		}

		if (oneregionmode)
			if (!region[curregion].sortvalid)
				for(i=256;i>0;i--)
				{
					koutp(97,kinp(97)|2);
					koutp(97,kinp(97)&~2);
				}

		for(i=numsplitlines-1;i>=0;i--)
		{
			drawline3d(&splitline[i],&splitline[i+1],254L);

			ox = splitline[i].x - ipos.x;
			oy = splitline[i].y - ipos.y;
			oz = splitline[i].z - ipos.z;
			x1 = (ox*istr.x + oy*istr.y + oz*istr.z)*fhxdim;
			y1 = (ox*ihei.x + oy*ihei.y + oz*ihei.z)*fhxdimyxaspect;
			z1 = ox*ifor.x + oy*ifor.y + oz*ifor.z;
			if (z1 >= 1)
			{
				r = 1/z1;
				ix = x1*r+fhxdim;
				iy = y1*r+fhydim;
				if ((ix >= 3) && (iy >= 3) && (ix < xdim-3) && (iy < ydim-3))
				{
					k = frameplace+ylookup[iy-3]+ix-3;
					for(y=2;y>=0;y--)
						for(x=2;x>=0;x--)
						{
							j = ((x<<1)-5)*((x<<1)-5) + ((y<<1)-5)*((y<<1)-5);
							if (j < 32)
							{
								*(char *)(k+ylookup[y]+x) = 254-(j<<2);
								*(char *)(k+ylookup[y]+(5-x)) = 254-(j<<2);
								*(char *)(k+ylookup[5-y]+x) = 254-(j<<2);
								*(char *)(k+ylookup[5-y]+(5-x)) = 254-(j<<2);
							}
						}
				}
			}
		}

			//Get search... data
		f = 1 / fhxdim;
		ox = (searchx-fhxdim)*f;
		oy = (searchy-fhydim)*f*fxyaspect;
		p0.x = ifor.x + ox*istr.x + oy*ihei.x;
		p0.y = ifor.y + ox*istr.y + oy*ihei.y;
		p0.z = ifor.z + ox*istr.z + oy*ihei.z;
		searchplane = hitscan(curregion, &ipos, &p0, &searchpos);
		if (searchplane >= 0) aligntogrid(&searchpos,searchplane,8L);

		//if (bstatus&2)
		//{
		//   i = frameplace+ylookup[searchy]+searchx;
		//   for(x=4;x>0;x--)
		//   {
		//      *(char *)(i+x) = 254;
		//      *(char *)(i-x) = 254;
		//      *(char *)(i+ylookup[x]) = 254;
		//      *(char *)(i-ylookup[x]) = 254;
		//   }
		//}
		//else
		//{
			i = frameplace+ylookup[searchy-3]+searchx-3;
			for(y=2;y>=0;y--)
				for(x=2;x>=0;x--)
				{
					j = ((x<<1)-5)*((x<<1)-5) + ((y<<1)-5)*((y<<1)-5);
					if (j < 32)
					{
						*(char *)(i+ylookup[y]+x) = 254-(j<<2);
						*(char *)(i+ylookup[y]+(5-x)) = 254-(j<<2);
						*(char *)(i+ylookup[5-y]+x) = 254-(j<<2);
						*(char *)(i+ylookup[5-y]+(5-x)) = 254-(j<<2);
					}
				}
		//}

		if (showgrid)
		{
			for(m=numregions-1;m>=0;m--)
			{
				if (!(gotregion[m>>3]&pow2char[m&7])) continue;
				p = startplane = region[m].planehead;
				do
				{
					i = edge[loop[plane[p].loophead].edgehead].v;
					if ((plane[p].o < 0) &&
						 (plane[p].normal.x*(ipos.x-vertex[i].x)
						 +plane[p].normal.y*(ipos.y-vertex[i].y)
						 +plane[p].normal.z*(ipos.z-vertex[i].z) > 0))
					{
						startloop = plane[p].loophead;
						l = startloop;
						do
						{
							startedge = loop[l].edgehead;
							e = startedge;
							do
							{
								i = edge[e].n;
								drawline3d(&vertex[edge[e].v],&vertex[edge[i].v],254L);

								ox = vertex[edge[e].v].x - ipos.x;
								oy = vertex[edge[e].v].y - ipos.y;
								oz = vertex[edge[e].v].z - ipos.z;
								x0 = (ox*istr.x + oy*istr.y + oz*istr.z)*fhxdim;
								y0 = (ox*ihei.x + oy*ihei.y + oz*ihei.z)*fhxdimyxaspect;
								z0 = ox*ifor.x + oy*ifor.y + oz*ifor.z;
								if (z0 >= 1)
								{
									z0 = 1/z0;
									ix = x0*z0+fhxdim;
									iy = y0*z0+fhydim;
									if ((ix >= 3) && (iy >= 3) && (ix < xdim-3) && (iy < ydim-3))
									{
										k = frameplace+ylookup[iy-3]+ix-3;
										for(y=2;y>=0;y--)
											for(x=2;x>=0;x--)
											{
												j = ((x<<1)-5)*((x<<1)-5) + ((y<<1)-5)*((y<<1)-5);
												if (j < 32)
												{
													*(char *)(k+ylookup[y]+x) = 254-(j<<2);
													*(char *)(k+ylookup[y]+(5-x)) = 254-(j<<2);
													*(char *)(k+ylookup[5-y]+x) = 254-(j<<2);
													*(char *)(k+ylookup[5-y]+(5-x)) = 254-(j<<2);
												}
											}
									}
								}

								e = i;
							} while (e != startedge);
							l = loop[l].n;
						} while (l != startloop);
					}
					p = plane[p].n;
				} while (p != startplane);
			}

			k = 0;
			startplane = j = region[curregion].planehead;
			do
			{
				i = 0;
				startloop = l = plane[j].loophead;
				do
				{
					startedge = e = loop[l].edgehead;
					do
					{
						p0.x = (vertex[edge[e].v].x + vertex[edge[edge[e].n].v].x)*.5;
						p0.y = (vertex[edge[e].v].y + vertex[edge[edge[e].n].v].y)*.5;
						p0.z = (vertex[edge[e].v].z + vertex[edge[edge[e].n].v].z)*.5;
						p1.x = (vertex[edge[edge[e].o].v].x + vertex[edge[edge[edge[e].o].n].v].x)*.5;
						p1.y = (vertex[edge[edge[e].o].v].y + vertex[edge[edge[edge[e].o].n].v].y)*.5;
						p1.z = (vertex[edge[edge[e].o].v].z + vertex[edge[edge[edge[e].o].n].v].z)*.5;
						if ((p0.x != p1.x) || (p0.y != p1.y) || (p0.z != p1.z))
						{
								//Draw midpoint tabs
							p1.x = vertex[edge[edge[e].n].v].x - vertex[edge[e].v].x;
							p1.y = vertex[edge[edge[e].n].v].y - vertex[edge[e].v].y;
							p1.z = vertex[edge[edge[e].n].v].z - vertex[edge[e].v].z;
							f = p1.x*p1.x + p1.y*p1.y + p1.z*p1.z;
							if (f > 0)
							{
								f = 4.0/sqrt(f);  //Normalize to 4 units long
								p1.x *= f; p1.y *= f; p1.z *= f;
								p2.x = p1.y*plane[j].normal.z - p1.z*plane[j].normal.y;
								p2.y = p1.z*plane[j].normal.x - p1.x*plane[j].normal.z;
								p2.z = p1.x*plane[j].normal.y - p1.y*plane[j].normal.x;
								p2.x += p0.x; p2.y += p0.y; p2.z += p0.z;
								drawline3d(&p0,&p2,0L);
							}
							for(x=16;x>0;x--)
							{
								koutp(97,kinp(97)|2);
								koutp(97,kinp(97)&~2);
							}
						}

						i++;
						e = edge[e].n;
					} while (e != startedge);
					l = loop[l].n;
				} while (l != startloop);

				sprintf(tempbuf,"%ld: %ld",k,i);
				if (j == searchplane)
					printsmall(0L,k*6+8,254L,-1L,tempbuf);
				else
					printsmall(0L,k*6+8,254L-64L,-1L,tempbuf);


				k++;

				j = plane[j].n;
			} while (j != startplane);

			sprintf(tempbuf,"Curregion = %ld",curregion);
			printsmall((xdim>>1)-(strlen(tempbuf)<<1),0L,254L,-1L,tempbuf);
		}

		if (vgacompatible)
		{
				//VR flag
			while (!(inp(0x3c2)&128));
			if (kinp(0x3c2)&128)
			{
				koutpw(0x3d4,((long)overtbits<<8)+0x11);
				koutp(0x3d5,overtbits+16);
			}
		}

		i = totalclock-ototalclock; ototalclock += i;
		j = ofinetotalclock-gettimer42(); ofinetotalclock -= j;
		i = 11931810/(((i*TIMERRATE-(j&65535)+32768)&0xffff0000)+(j&65535));
		frameval[framecnt&(AVERAGEFRAMES-1)] = i; framecnt++;
			//Print MAX FRAME RATE
		i = frameval[0];
		for(j=AVERAGEFRAMES-1;j>0;j--) i = max(i,frameval[j]);
		averagefps = ((averagefps*3+i)>>2);
		sprintf(tempbuf,"%ld.%ld",averagefps/10,averagefps%10);
		printsmall(0L,0L,254L,-1L,tempbuf);

		visualpage = activepage;
		setvisualpage(visualpage); //|0x80000000);
		activepage++; if (activepage >= maxpages) activepage = 0;

		if (totalclock >= lastreadmouse+4)
		{
			lastreadmouse += 4;

			ang1vel *= .8;
			ang2vel *= .8;
			ang3vel *= .8;
			if (fabs(ang1vel) < .5) ang1vel = 0;
			if (fabs(ang2vel) < .5) ang2vel = 0;
			if (fabs(ang3vel) < .5) ang3vel = 0;

			obstatus = bstatus;
			readmousexy(&mousx,&mousy);
			readmousebstatus(&bstatus);

			if (!(bstatus&2)) ang1vel += mousx; else ang3vel -= mousx;
			ang2vel -= mousy;
			if (ang2 < -PI*.75) ang2 += ((-PI*.75)-ang2)*.125;
			if (ang2 > PI*.75) ang2 += ((PI*.75)-ang2)*.125;

			while (ang3 < -PI) ang3 += PI2;
			while (ang3 > PI) ang3 -= PI2;
			ang3vel -= ang3*16;
		}
		ang1 += (float)ang1vel*(float)synctics*(1.75/8192.0);
		ang2 += (float)ang2vel*(float)synctics*(1.75/8192.0);
		ang3 += (float)ang3vel*(float)synctics*(1.75/8192.0);
		setuprotate3d();

		if (keystatus[0x9d])  //Rt.ctrl
		{
			ipos.x -= ihei.x*synctics*.5;
			ipos.y -= ihei.y*synctics*.5;
			ipos.z -= ihei.z*synctics*.5;
		}
		if (keystatus[0x52])  //Keyp.0
		{
			ipos.x += ihei.x*synctics*.5;
			ipos.y += ihei.y*synctics*.5;
			ipos.z += ihei.z*synctics*.5;
		}
		if (keystatus[0xc8])
		{
			ipos.x += ifor.x*synctics*.5;
			ipos.y += ifor.y*synctics*.5;
			ipos.z += ifor.z*synctics*.5;
		}
		if (keystatus[0xd0])
		{
			ipos.x -= ifor.x*synctics*.5;
			ipos.y -= ifor.y*synctics*.5;
			ipos.z -= ifor.z*synctics*.5;
		}
		if (keystatus[0xcb])
		{
			ipos.x -= istr.x*synctics*.5;
			ipos.y -= istr.y*synctics*.5;
			ipos.z -= istr.z*synctics*.5;
		}
		if (keystatus[0xcd])
		{
			ipos.x += istr.x*synctics*.5;
			ipos.y += istr.y*synctics*.5;
			ipos.z += istr.z*synctics*.5;
		}
		if (keystatus[0x9c])
		{
			keystatus[0x9c] = 0;
			gtexturemode++;
			if (gtexturemode >= 2) gtexturemode = 0;
		}

		if (keystatus[0x28])
		{
			keystatus[0x28] = 0;
			for(i=numvertices-1;i>=0;i--)
			{
				vertex[i].x += (float)((rand()&255)-128)/32;
				vertex[i].y += (float)((rand()&255)-128)/32;
				vertex[i].z += (float)((rand()&255)-128)/32;
			}
		}

		if (keystatus[0xf]) //Tab
		{
			keystatus[0xf] = 0;
			if (searchplane >= 0)
				temppicnum = plane[searchplane].picnum;
		}
		if (keystatus[0x1c]) //L.Enter
		{
			keystatus[0x1c] = 0;
			if ((searchplane >= 0) && (temppicnum >= 0))
				plane[searchplane].picnum = temppicnum;
		}
		if (keystatus[0x4a])
		{
			keystatus[0x4a] = 0;
			if (searchplane >= 0)
				plane[searchplane].picnum = max(plane[searchplane].picnum-1,0);
		}
		if (keystatus[0x4e])
		{
			keystatus[0x4e] = 0;
			if (searchplane >= 0)
				plane[searchplane].picnum++;
		}

		if (keystatus[0x2b])  // Backslash
		{
			keystatus[0x2b] = 0;
			oneregionmode ^= 1;
		}
		if ((keystatus[0x35]) && (searchplane >= 0))
		{
			keystatus[0x35] = 0;
			calcplanetexturedefault(searchplane);
		}
		if ((keystatus[0x4b]) && (searchplane >= 0))   //Keyp. 4
		{
			if (!(keystatus[0x2a]|keystatus[0x36]))
			{
				if (keystatus[0x4c])
				{
					plane[searchplane].uscale -= 8;
					plane[searchplane].uscale += 4;
					plane[searchplane].uscale &= 0xfff8;
					keystatus[0x4b] = 0;
				}
				else
					plane[searchplane].uscale--;
			}
			else
			{
				if (keystatus[0x4c])
				{
					plane[searchplane].uoffs -= 8192;
					plane[searchplane].uoffs += 4096;
					plane[searchplane].uoffs &= 0xe000;
					keystatus[0x4b] = 0;
				}
				else
					plane[searchplane].uoffs -= 256;
			}
		}
		if ((keystatus[0x4d]) && (searchplane >= 0))   //Keyp. 6
		{
			if (!(keystatus[0x2a]|keystatus[0x36]))
			{
				if (keystatus[0x4c])
				{
					plane[searchplane].uscale += 8;
					plane[searchplane].uscale += 4;
					plane[searchplane].uscale &= 0xfff8;
					keystatus[0x4d] = 0;
				}
				else
					plane[searchplane].uscale++;
			}
			else
			{
				if (keystatus[0x4c])
				{
					plane[searchplane].uoffs += 8192;
					plane[searchplane].uoffs += 4096;
					plane[searchplane].uoffs &= 0xe000;
					keystatus[0x4d] = 0;
				}
				else
					plane[searchplane].uoffs += 256;
			}
		}
		if ((keystatus[0x48]) && (searchplane >= 0))  //Keyp. 8
		{
			if (!(keystatus[0x2a]|keystatus[0x36]))
			{
				if (keystatus[0x4c])
				{
					plane[searchplane].vscale -= 8;
					plane[searchplane].vscale += 4;
					plane[searchplane].vscale &= 0xfff8;
					keystatus[0x48] = 0;
				}
				else
					plane[searchplane].vscale--;
			}
			else
			{
				if (keystatus[0x4c])
				{
					plane[searchplane].voffs -= 8192;
					plane[searchplane].voffs += 4096;
					plane[searchplane].voffs &= 0xe000;
					keystatus[0x48] = 0;
				}
				else
					plane[searchplane].voffs -= 256;
			}
		}
		if ((keystatus[0x50]) && (searchplane >= 0))  //Keyp. 2
		{
			if (!(keystatus[0x2a]|keystatus[0x36]))
			{
				if (keystatus[0x4c])
				{
					plane[searchplane].vscale += 8;
					plane[searchplane].vscale += 4;
					plane[searchplane].vscale &= 0xfff8;
					keystatus[0x50] = 0;
				}
				else
					plane[searchplane].vscale++;
			}
			else
			{
				if (keystatus[0x4c])
				{
					plane[searchplane].voffs += 8192;
					plane[searchplane].voffs += 4096;
					plane[searchplane].voffs &= 0xe000;
					keystatus[0x50] = 0;
				}
				else
					plane[searchplane].voffs += 256;
			}
		}

		if (keystatus[0x13])  //R - removecollinears
		{
			keystatus[0x13] = 0;
			if (searchplane >= 0)
				removecollinears(searchplane);
		}
		if (keystatus[0x26])  //L - gridlock
		{
			keystatus[0x26] = 0;
			gridlock ^= 1;
		}
		if (keystatus[0x22])  //G - showgrid
		{
			keystatus[0x22] = 0;
			showgrid ^= 1;
		}

		if (keystatus[0x31])  //N - new board
		{
			keystatus[0x31] = 0;
			loadboard(boardfilename);
		}

		if (keystatus[0x21])  //F - front a poly
		{
			keystatus[0x21] = 0;
			if (keystatus[0x38]|keystatus[0xb8])   //ALT-F  Flip first point
			{
				if (searchplane >= 0)
				{
					if (keystatus[0x2a]|keystatus[0x36])  //Sh.Alt.F - Flip loop
						plane[searchplane].loophead = loop[plane[searchplane].loophead].n;
					else
						loop[plane[searchplane].loophead].edgehead = edge[loop[plane[searchplane].loophead].edgehead].n;
					calcplanetexture(searchplane);
				}
			}
			else if (keystatus[0x2a]|keystatus[0x36])  //SHIFT-F Change align method
			{
				if (searchplane >= 0)
					plane[searchplane].alignmethod = ((plane[searchplane].alignmethod+1)&3);
			}
			else
			{
				if (searchplane >= 0)
				{
					if (searchplane != region[plane[searchplane].region].planehead)
					{
						plane[lastplane(searchplane)].n = plane[searchplane].n;
						plane[searchplane].n = region[plane[searchplane].region].planehead;
						plane[lastplane(region[plane[searchplane].region].planehead)].n = searchplane;
						region[plane[searchplane].region].planehead = searchplane;
					}
				}
			}
		}
		if (keystatus[0x30])  //B - back a poly
		{
			keystatus[0x30] = 0;
			if (searchplane >= 0)
			{
				if (searchplane == region[plane[searchplane].region].planehead)
					region[plane[searchplane].region].planehead = plane[searchplane].n;
				else
				{
					plane[lastplane(searchplane)].n = plane[searchplane].n;
					plane[searchplane].n = region[plane[searchplane].region].planehead;
					plane[lastplane(region[plane[searchplane].region].planehead)].n = searchplane;
				}
			}
		}

		if (keystatus[0x24])  //J (join plane)
		{
			keystatus[0x24] = 0;
			if (searchplane >= 0)
			{
				i = edge[loop[plane[joinplane].loophead].edgehead].v;
				j = edge[loop[plane[searchplane].loophead].edgehead].v;
				if (joinplane < 0)
					joinplane = searchplane;
				else if (plane[joinplane].region != plane[searchplane].region)
				{
						//Join joinplane's region to searchplane's region

						//Delete planes that share both regions
					for(i=numplanes-1;i>=0;i--)
					{
						if (plane[i].o < 0) continue;

						if ((plane[i].region == plane[searchplane].region) &&
							 (plane[plane[i].o].region == plane[joinplane].region))
						{

								//Delete all loops & edges of plane plane[i].o
							l = startloop = plane[plane[i].o].loophead;
							do
							{
								ll = loop[l].n;
								e = startedge = loop[l].edgehead;
								do
								{
									ee = edge[e].n;

										//Delete edge e
									numedges--;
									memcpy(&edge[e],&edge[numedges],sizeof(edgetype));
									for(j=numedges-1;j>=0;j--)
									{
										if (edge[j].n == numedges) edge[j].n = e;
										if (edge[j].l == numedges) edge[j].l = e;
										if (edge[j].o == numedges) edge[j].o = e;
									}
									for(j=numloops-1;j>=0;j--)
										if (loop[j].edgehead == numedges)
											loop[j].edgehead = e;
									if (ee == numedges) ee = e;

									e = ee;
								} while (e != startedge);

									//Delete loop l
								numloops--;
								memcpy(&loop[l],&loop[numloops],sizeof(looptype));
								for(j=numedges-1;j>=0;j--)
									if (edge[j].loop == numloops)
										edge[j].loop = l;
								for(j=numloops-1;j>=0;j--)
									if (loop[j].n == numloops)
										loop[j].n = l;
								for(j=numplanes-1;j>=0;j--)
									if (plane[j].loophead == numloops)
										plane[j].loophead = l;
								if (ll == numloops) ll = l;

								l = ll;
							} while (l != startloop);

								//Delete all loops & edges of plane i
							l = startloop = plane[i].loophead;
							do
							{
								ll = loop[l].n;
								e = startedge = loop[l].edgehead;
								do
								{
									ee = edge[e].n;

										//Delete edge e
									numedges--;
									memcpy(&edge[e],&edge[numedges],sizeof(edgetype));
									for(j=numedges-1;j>=0;j--)
									{
										if (edge[j].n == numedges) edge[j].n = e;
										if (edge[j].l == numedges) edge[j].l = e;
										if (edge[j].o == numedges) edge[j].o = e;
									}
									for(j=numloops-1;j>=0;j--)
										if (loop[j].edgehead == numedges)
											loop[j].edgehead = e;
									if (ee == numedges) ee = e;

									e = ee;
								} while (e != startedge);

									//Delete loop l
								numloops--;
								memcpy(&loop[l],&loop[numloops],sizeof(looptype));
								for(j=numedges-1;j>=0;j--)
									if (edge[j].loop == numloops)
										edge[j].loop = l;
								for(j=numloops-1;j>=0;j--)
									if (loop[j].n == numloops)
										loop[j].n = l;
								for(j=numplanes-1;j>=0;j--)
									if (plane[j].loophead == numloops)
										plane[j].loophead = l;
								if (ll == numloops) ll = l;

								l = ll;
							} while (l != startloop);

							k = plane[i].o; l = i;

								//Re-wire searchplane out of region
							for(j=numregions-1;j>=0;j--)
								if (region[j].planehead == k)
									region[j].planehead = plane[k].n;
							for(j=numplanes-1;j>=0;j--)
								if (plane[j].n == k)
									plane[j].n = plane[k].n;

								//Delete plane[i].o
							numplanes--;
							memcpy(&plane[k],&plane[numplanes],sizeof(planetype));
							for(j=numloops-1;j>=0;j--)
								if (loop[j].plane == numplanes)
									loop[j].plane = k;
							for(j=numplanes-1;j>=0;j--)
							{
								if (plane[j].n == numplanes) plane[j].n = k;
								if (plane[j].o == numplanes) plane[j].o = k;
							}
							for(j=numregions-1;j>=0;j--)
								if (region[j].planehead == numplanes)
									region[j].planehead = k;
							if (searchplane == numplanes) searchplane = k;
							if (joinplane == numplanes) joinplane = k;
							if (l == numplanes) l = k;

								//Re-wire searchplane out of region
							for(j=numregions-1;j>=0;j--)
								if (region[j].planehead == l)
									region[j].planehead = plane[l].n;
							for(j=numplanes-1;j>=0;j--)
								if (plane[j].n == l)
									plane[j].n = plane[l].n;

								//Delete plane[i]
							numplanes--;
							memcpy(&plane[l],&plane[numplanes],sizeof(planetype));
							for(j=numloops-1;j>=0;j--)
								if (loop[j].plane == numplanes)
									loop[j].plane = l;
							for(j=numplanes-1;j>=0;j--)
							{
								if (plane[j].n == numplanes) plane[j].n = l;
								if (plane[j].o == numplanes) plane[j].o = l;
							}
							for(j=numregions-1;j>=0;j--)
								if (region[j].planehead == numplanes)
									region[j].planehead = l;
							if (searchplane == numplanes) searchplane = l;
							if (joinplane == numplanes) joinplane = l;
						}
					}

						//Re-wire searchplane's planes into joinplane's regions
					i = plane[searchplane].n;
					j = plane[joinplane].n;
					plane[joinplane].n = i;
					plane[searchplane].n = j;

						//Delete searchplane's region
					j = plane[searchplane].region;
					numregions--;
					memcpy(&region[j],&region[numregions],sizeof(regiontype));
					for(i=numplanes-1;i>=0;i--)
						if (plane[i].region == j)
							plane[i].region = plane[joinplane].region;

						//SLOW way to fix edge opposites...
					k = plane[joinplane].region;
					for(i=0;i<numedges;i++)
					{
						if (plane[loop[edge[i].loop].plane].region != k) continue;
						if ((edge[i].v == edge[edge[edge[i].o].n].v) && (edge[edge[i].o].v == edge[edge[i].n].v))
							continue;
						for(j=0;j<numedges;j++)
							if ((edge[i].v == edge[edge[j].n].v) && (edge[j].v == edge[edge[i].n].v))
								if (plane[loop[edge[j].loop].plane].region == k)
									{ edge[i].o = j; edge[j].o = i; break; }
					}

					regionsort(plane[joinplane].region);
					joinplane = searchplane = -1;
				}
				else if ((joinplane != searchplane) &&
					(fabs(plane[joinplane].normal.x-plane[searchplane].normal.x) < .0001) &&
					(fabs(plane[joinplane].normal.y-plane[searchplane].normal.y) < .0001) &&
					(fabs(plane[joinplane].normal.z-plane[searchplane].normal.z) < .0001) &&
					((vertex[i].x-vertex[j].x)*plane[searchplane].normal.x +
					 (vertex[i].y-vertex[j].y)*plane[searchplane].normal.y +
					 (vertex[i].z-vertex[j].z)*plane[searchplane].normal.z < .0001))
				{
						//Join joinplane to searchplane

						//Delete edges that border both planes
					for(i=numedges-1;i>=0;i--)
					{
						if ((loop[edge[i].loop].plane != joinplane) || (loop[edge[edge[i].o].loop].plane != searchplane))
							continue;

						if (edge[edge[i].l].loop != edge[edge[edge[i].o].n].loop)
						{
							k = edge[edge[edge[i].o].n].loop;

								//Re-wire loop k out of list
							for(j=numedges-1;j>=0;j--)
								if (edge[j].loop == k)
									edge[j].loop = edge[edge[i].l].loop;
							for(j=numloops-1;j>=0;j--)
								if (loop[j].n == k)
									loop[j].n = loop[k].n;
							for(j=numplanes-1;j>=0;j--)
								if (plane[j].loophead == k)
									plane[j].loophead = loop[k].n;

							numloops--;
							memcpy(&loop[k],&loop[numloops],sizeof(looptype));
							for(j=numedges-1;j>=0;j--)
								if (edge[j].loop == numloops)
									edge[j].loop = k;
							for(j=numloops-1;j>=0;j--)
								if (loop[j].n == numloops)
									loop[j].n = k;
							for(j=numplanes-1;j>=0;j--)
								if (plane[j].loophead == numloops)
									plane[j].loophead = k;
						}

						if (edge[edge[i].n].loop != edge[edge[edge[i].o].l].loop)
						{
							k = edge[edge[edge[i].o].l].loop;

								//Re-wire loop k out of list
							for(j=numedges-1;j>=0;j--)
								if (edge[j].loop == k)
									edge[j].loop = edge[edge[i].n].loop;
							for(j=numloops-1;j>=0;j--)
								if (loop[j].n == k)
									loop[j].n = loop[k].n;
							for(j=numplanes-1;j>=0;j--)
								if (plane[j].loophead == k)
									plane[j].loophead = loop[k].n;

							numloops--;
							memcpy(&loop[k],&loop[numloops],sizeof(looptype));
							for(j=numedges-1;j>=0;j--)
								if (edge[j].loop == numloops)
									edge[j].loop = k;
							for(j=numloops-1;j>=0;j--)
								if (loop[j].n == numloops)
									loop[j].n = k;
							for(j=numplanes-1;j>=0;j--)
								if (plane[j].loophead == numloops)
									plane[j].loophead = k;
						}

						edge[edge[i].l].n = edge[edge[i].o].n;
						edge[edge[i].n].l = edge[edge[i].o].l;
						edge[edge[edge[i].o].n].l = edge[i].l;
						edge[edge[edge[i].o].l].n = edge[i].n;

						k = i; l = edge[i].o;
							//Delete edge edge[i].o here!
						numedges--;
						memcpy(&edge[k],&edge[numedges],sizeof(edgetype));
						for(j=numedges-1;j>=0;j--)
						{
							if (edge[j].n == numedges) edge[j].n = k;
							if (edge[j].l == numedges) edge[j].l = k;
							if (edge[j].o == numedges) edge[j].o = k;
						}
						for(j=numloops-1;j>=0;j--)
							if (loop[j].edgehead == numedges)
								loop[j].edgehead = k;

							//I hate ugly cases like this
							//(If the edge just copied from numedges gets deleted)
						if (l == numedges) l = k;

							//Delete edge i here!
						numedges--;
						memcpy(&edge[l],&edge[numedges],sizeof(edgetype));
						for(j=numedges-1;j>=0;j--)
						{
							if (edge[j].n == numedges) edge[j].n = l;
							if (edge[j].l == numedges) edge[j].l = l;
							if (edge[j].o == numedges) edge[j].o = l;
						}
						for(j=numloops-1;j>=0;j--)
							if (loop[j].edgehead == numedges)
								loop[j].edgehead = l;
					}

						//Copy searchplane's loops to joinplane (not finished yet)
					//loop[plane[searchplane].loophead].n = plane[searchplane].loophead;
					//loop[plane[joinplane].loophead].n = plane[searchplane].loophead;

						//Re-wire searchplane out of region
					for(i=numregions-1;i>=0;i--)
						if (region[i].planehead == searchplane)
							region[i].planehead = plane[searchplane].n;
					for(i=numplanes-1;i>=0;i--)
					{
						if (plane[i].n == searchplane) plane[i].n = plane[searchplane].n;
						if (plane[i].o == searchplane) plane[i].o = joinplane;
					}

						//Delete searchplane
					numplanes--;
					memcpy(&plane[searchplane],&plane[numplanes],sizeof(planetype));
					for(i=numloops-1;i>=0;i--)
						if (loop[i].plane == numplanes)
							loop[i].plane = searchplane;
					for(i=numplanes-1;i>=0;i--)
					{
						if (plane[i].n == numplanes) plane[i].n = searchplane;
						if (plane[i].o == numplanes) plane[i].o = searchplane;
					}
					for(i=numregions-1;i>=0;i--)
						if (region[i].planehead == numplanes)
							region[i].planehead = searchplane;
					if (joinplane == numplanes) joinplane = searchplane;

					removecollinears(joinplane);

					cleanvertexlist();

					calcplanetexturedefault(joinplane);

					regionsort(plane[joinplane].region);
					joinplane = searchplane = -1;
				}
			}
		}

		if (keystatus[0xe])
		{
			keystatus[0xe] = 0;
			if (numsplitlines > 0) numsplitlines--;
		}
		if (keystatus[0x39])
		{
			keystatus[0x39] = 0;
			if (searchplane >= 0)
			{
				if ((splitmode >= 0) && (numsplitlines >= 1) && (getedgeofvertexonplane(&searchpos,splitplane) >= 0))
				{
						//Split polygon
					numsplitlines++;

					k = getedgeofvertexonplane(&searchpos,splitplane);
						//edge[k] to edge[splitmode]

						//If split has beginning & end on same line, insert fake point
					if (k == splitmode)
					{
						j = getvertex((splitline[0].x+splitline[numsplitlines-1].x)*.5,
										  (splitline[0].y+splitline[numsplitlines-1].y)*.5,
										  (splitline[0].z+splitline[numsplitlines-1].z)*.5);
						edge[numedges].v = j;
						edge[numedges].n = edge[splitmode].n;
						edge[numedges].l = splitmode;
						edge[numedges].o = numedges+1;
						edge[numedges].loop = edge[splitmode].loop;
						edge[numedges+1].v = edge[edge[splitmode].o].v;
						edge[numedges+1].n = edge[splitmode].o;
						edge[numedges+1].l = edge[edge[splitmode].o].l;
						edge[numedges+1].o = numedges;
						edge[numedges+1].loop = edge[edge[splitmode].o].loop;
						edge[edge[splitmode].n].l = numedges;
						edge[splitmode].n = numedges;
						edge[edge[edge[splitmode].o].l].n = numedges+1;
						edge[edge[splitmode].o].l = numedges+1;
						edge[edge[splitmode].o].v = j;
						numedges += 2;

						splitmode = getedgeofvertexonplane(&splitline[0],splitplane);
						k = getedgeofvertexonplane(&searchpos,splitplane);
					}

					plane[numplanes].loophead = numloops;
					plane[numplanes].n = plane[splitplane].n;
					plane[numplanes].o = -1;
					plane[numplanes].region = plane[splitplane].region;
					plane[splitplane].n = numplanes;

					loop[plane[splitplane].loophead].edgehead = splitmode;
					loop[numloops].edgehead = numedges+(numsplitlines<<1)-1;
					loop[numloops].n = numloops;
					loop[numloops].plane = numplanes;

					for(i=0;i<numsplitlines;i++)
					{
						j = getvertex(splitline[i].x,splitline[i].y,splitline[i].z);

						edge[numedges+i].v = j;
						edge[numedges+i].n = numedges+i+1;
						edge[numedges+i].l = numedges+i-1;
						edge[numedges+i].o = numedges+(numsplitlines<<1)-i-2;
						edge[numedges+i].loop = edge[splitmode].loop;
						edge[numedges+(numsplitlines<<1)-i-1].v = j;
						edge[numedges+(numsplitlines<<1)-i-1].n = numedges+(numsplitlines<<1)-i;
						edge[numedges+(numsplitlines<<1)-i-1].l = numedges+(numsplitlines<<1)-i-2;
						edge[numedges+(numsplitlines<<1)-i-1].o = numedges+i-1;
						edge[numedges+(numsplitlines<<1)-i-1].loop = edge[k].loop;
					}

					//Beware of using these variables later - they get destroyed!
					//oldedgesplitmoden = edge[splitmode].n;
					//oldedgekn = edge[k].n;

						//Connect new splitline edges to the 2 respective planes
					edge[edge[splitmode].n].l = numedges+(numsplitlines<<1)-1;
					edge[numedges+(numsplitlines<<1)-1].n = edge[splitmode].n;
					edge[edge[k].n].l = numedges+numsplitlines-1;
					edge[numedges+numsplitlines-1].n = edge[k].n;
					edge[splitmode].n = numedges;
					edge[numedges].l = splitmode;
					edge[k].n = numedges+numsplitlines;
					edge[numedges+numsplitlines].l = k;

						//Insert edges on adjacent planes for intersection points
					edge[numedges+(numsplitlines<<1)].v = edge[edge[splitmode].o].v;
					edge[numedges+(numsplitlines<<1)].n = edge[splitmode].o;
					edge[numedges+(numsplitlines<<1)].l = edge[edge[splitmode].o].l;
					edge[numedges+(numsplitlines<<1)].o = numedges+(numsplitlines<<1)-1;
					edge[numedges+(numsplitlines<<1)].loop = edge[edge[splitmode].o].loop;
					edge[edge[edge[splitmode].o].l].n = numedges+(numsplitlines<<1);
					edge[edge[splitmode].o].l = numedges+(numsplitlines<<1);
					edge[numedges+(numsplitlines<<1)-1].o = numedges+(numsplitlines<<1);
					edge[edge[splitmode].o].v = edge[numedges].v;

					edge[numedges+(numsplitlines<<1)+1].v = edge[edge[k].o].v;
					edge[numedges+(numsplitlines<<1)+1].n = edge[k].o;
					edge[numedges+(numsplitlines<<1)+1].l = edge[edge[k].o].l;
					edge[numedges+(numsplitlines<<1)+1].o = numedges+numsplitlines-1;
					edge[numedges+(numsplitlines<<1)+1].loop = edge[edge[k].o].loop;
					edge[edge[edge[k].o].l].n = numedges+(numsplitlines<<1)+1;
					edge[edge[k].o].l = numedges+(numsplitlines<<1)+1;
					edge[numedges+numsplitlines-1].o = numedges+(numsplitlines<<1)+1;
					edge[edge[k].o].v = edge[numedges+numsplitlines].v;

						//Danger of adjacent planes having collinear texture alignment
					j = edge[edge[splitmode].o].loop;
					if (loop[j].edgehead == edge[splitmode].o)
						loop[j].edgehead = edge[loop[j].edgehead].l;
					calcplanetexture(loop[edge[edge[splitmode].o].loop].plane);

					j = edge[edge[k].o].loop;
					if (loop[j].edgehead == edge[k].o)
						loop[j].edgehead = edge[loop[j].edgehead].l;
					calcplanetexture(loop[edge[edge[k].o].loop].plane);

					numvertices += numsplitlines;
					numedges += (numsplitlines<<1)+2;  //2 extra for the adjacent planes
					numloops++;
					numplanes++;

						//Fix edge[].loop's for splitplane
					startloop = l = plane[splitplane].loophead;
					do
					{
						startedge = e = loop[l].edgehead;
						do
						{
							edge[e].loop = l;
							e = edge[e].n;
						} while (e != startedge);
						l = loop[l].n;
					} while (l != startloop);

						//Fix edge[].p's for numplanes-1
					startloop = l = plane[numplanes-1].loophead;
					do
					{
						startedge = e = loop[l].edgehead;
						do
						{
							edge[e].loop = l;
							e = edge[e].n;
						} while (e != startedge);
						l = loop[l].n;
					} while (l != startloop);


					removecollinears(splitplane);
					removecollinears(numplanes-1);

					calcplanenormal(numplanes-1);
					calcplanetexturedefault(numplanes-1);
					calcplanetexturedefault(splitplane);

					regionsort(plane[splitplane].region);
					numsplitlines = 0;
				}
				else if ((searchpos.x == splitline[0].x) && (searchpos.y == splitline[0].y) && (searchpos.z == splitline[0].z) && (numsplitlines >= 3))
				{
						//Insert loop into polygon
					plane[numplanes].loophead = numloops+1;
					plane[numplanes].n = plane[splitplane].n;
					plane[numplanes].o = -1;
					plane[numplanes].region = plane[splitplane].region;
					plane[splitplane].n = numplanes;

					loop[numloops].edgehead = numedges;
					loop[numloops].n = loop[plane[splitplane].loophead].n;
					loop[numloops].plane = splitplane;
					loop[plane[splitplane].loophead].n = numloops;
					loop[numloops+1].edgehead = numedges+numsplitlines;
					loop[numloops+1].n = numloops+1;
					loop[numloops+1].plane = numplanes;

					for(i=0;i<numsplitlines;i++)
					{
						k = getvertex(splitline[i].x,splitline[i].y,splitline[i].z);

						edge[numedges+i].v = k;
						j = i+1; if (j >= numsplitlines) j = 0;
						edge[numedges+i].n = numedges+j;
						j = i-1; if (j < 0) j = numsplitlines-1;
						edge[numedges+i].l = numedges+j;
						edge[numedges+i].o = numedges+(numsplitlines<<1)-1-i;
						edge[numedges+i].loop = numloops;

						l = (numsplitlines<<1)-i;
						if (l >= (numsplitlines<<1)) l -= numsplitlines;

						edge[numedges+l].v = k;
						j = l+1; if (j >= (numsplitlines<<1)) j = numsplitlines;
						edge[numedges+l].n = numedges+j;
						j = l-1; if (j < numsplitlines) j = (numsplitlines<<1)-1;
						edge[numedges+l].l = numedges+j;
						edge[numedges+l].o = numedges+(numsplitlines<<1)-1-l;
						edge[numedges+l].loop = numloops+1;
					}
					numedges += (numsplitlines<<1);
					numloops += 2;
					numplanes++;

					calcplanenormal(numplanes-1);

						//Swap loops if drawn in wrong direction
					if ((fabs(plane[splitplane].normal.x+plane[numplanes-1].normal.x)+
						  fabs(plane[splitplane].normal.y+plane[numplanes-1].normal.y)+
						  fabs(plane[splitplane].normal.z+plane[numplanes-1].normal.z) < 0.001))
					{
						for(i=numsplitlines-1;i>=0;i--)
							swaplong(&edge[numedges+i-(numsplitlines<<1)].v,&edge[numedges+i-numsplitlines].v);
						plane[numplanes-1].normal.x = plane[splitplane].normal.x;
						plane[numplanes-1].normal.y = plane[splitplane].normal.y;
						plane[numplanes-1].normal.z = plane[splitplane].normal.z;
					}

					calcplanetexturedefault(numplanes-1);

					regionsort(plane[splitplane].region);
					numsplitlines = 0;
				}
				else
				{
					if (!numsplitlines)
					{
						splitplane = searchplane;
						splitmode = getedgeofvertexonplane(&searchpos,splitplane);
						splitline[numsplitlines] = searchpos;
					}
					numsplitlines++;
				}
			}
		}
		if ((numsplitlines > 0) && (searchplane == splitplane))
			splitline[numsplitlines] = searchpos;

		dir = 0;
		if (keystatus[0xc9]) { keystatus[0xc9] = 0; dir = 4; }  //PGUP
		if (keystatus[0xd1]) { keystatus[0xd1] = 0; dir = -4; } //PGDN
		if ((dir) && (searchplane >= 0)) //PGUP/PGDN
		{
			removecollinears(searchplane);

				//Step 1:  If side planes are parallel, then extend existing plane
				//         Otherwise create a new side plane with 4 sides
			pla = (planetype *)&plane[searchplane];
			startloop = l = pla->loophead;
			do
			{
				startedge = e = loop[l].edgehead;
				do
				{
					ee = edge[e].n;

						//Calculate normal of side plane
					x0 = vertex[edge[ee].v].x-vertex[edge[e].v].x;
					y0 = vertex[edge[ee].v].y-vertex[edge[e].v].y;
					z0 = vertex[edge[ee].v].z-vertex[edge[e].v].z;
					x1 = plane[searchplane].normal.x;
					y1 = plane[searchplane].normal.y;
					z1 = plane[searchplane].normal.z;
					ox = y0*z1 - z0*y1;
					oy = z0*x1 - x0*z1;
					oz = x0*y1 - y0*x1;
					f = sqrt(ox*ox + oy*oy + oz*oz); if (f) f = 1 / f;
					ox *= f; oy *= f; oz *= f;

					i = loop[edge[edge[e].o].loop].plane;
						//If normals are parallel, then just modify existing plane
					if ((fabs(plane[i].normal.x-ox)+fabs(plane[i].normal.y-oy)+fabs(plane[i].normal.z-oz) < 0.001) ||
						 (fabs(plane[i].normal.x+ox)+fabs(plane[i].normal.y+oy)+fabs(plane[i].normal.z+oz) < 0.001))

					{
						i = edge[e].o; j = edge[i].n;

							//Collinear test for 1st point
						x0 = vertex[edge[i].v].x-vertex[edge[edge[i].l].v].x;
						y0 = vertex[edge[i].v].y-vertex[edge[edge[i].l].v].y;
						z0 = vertex[edge[i].v].z-vertex[edge[edge[i].l].v].z;
						x1 = pla->normal.x;
						y1 = pla->normal.y;
						z1 = pla->normal.z;
						if (fabs(y0*z1-z0*y1)+fabs(z0*x1-x0*z1)+fabs(x0*y1-y0*x1) < 0.001)
						{
								//Drag existing point on existing side plane
							edge[i].v = getvertex(vertex[edge[ee].v].x + pla->normal.x*dir,
														 vertex[edge[ee].v].y + pla->normal.y*dir,
														 vertex[edge[ee].v].z + pla->normal.z*dir);
						}
						else
						{
								//Insert new point on existing side plane
							edge[numedges].v = getvertex(vertex[edge[ee].v].x + pla->normal.x*dir,
																  vertex[edge[ee].v].y + pla->normal.y*dir,
																  vertex[edge[ee].v].z + pla->normal.z*dir);

							edge[numedges].n = edge[i].n;
							edge[numedges].l = i;
							edge[numedges].loop = edge[i].loop;

							edge[edge[i].n].l = numedges;
							edge[i].n = numedges;
							numedges++;
						}

							//Collinear test for 2nd point
						x0 = vertex[edge[j].v].x-vertex[edge[edge[j].n].v].x;
						y0 = vertex[edge[j].v].y-vertex[edge[edge[j].n].v].y;
						z0 = vertex[edge[j].v].z-vertex[edge[edge[j].n].v].z;
						x1 = pla->normal.x;
						y1 = pla->normal.y;
						z1 = pla->normal.z;
						if (fabs(y0*z1-z0*y1)+fabs(z0*x1-x0*z1)+fabs(x0*y1-y0*x1) < 0.001)
						{
								//Drag existing point on existing side plane
							edge[j].v = getvertex(vertex[edge[e].v].x + pla->normal.x*dir,
														 vertex[edge[e].v].y + pla->normal.y*dir,
														 vertex[edge[e].v].z + pla->normal.z*dir);
						}
						else
						{
								//Insert new point on existing side plane
							edge[numedges].v = getvertex(vertex[edge[e].v].x + pla->normal.x*dir,
																  vertex[edge[e].v].y + pla->normal.y*dir,
																  vertex[edge[e].v].z + pla->normal.z*dir);

							edge[numedges].n = j;
							edge[numedges].l = edge[j].l;
							edge[numedges].loop = edge[j].loop;

							edge[edge[j].l].n = numedges;
							edge[j].l = numedges;
							numedges++;
						}

						calcplanetexturedefault(loop[edge[edge[e].o].loop].plane);
						calcplanenormal(loop[edge[edge[e].o].loop].plane);
					}
					else
					{
							//Create new side plane with 4 sides
						plane[numplanes].loophead = numloops;
						plane[numplanes].n = plane[searchplane].n;
						plane[numplanes].o = -1;
						plane[numplanes].region = plane[searchplane].region;

						loop[numloops].edgehead = numedges;
						loop[numloops].n = numloops;
						loop[numloops].plane = numplanes;

						edge[numedges  ].v = edge[ee].v;
						edge[numedges+1].v = getvertex(vertex[edge[ee].v].x + pla->normal.x*dir,
																 vertex[edge[ee].v].y + pla->normal.y*dir,
																 vertex[edge[ee].v].z + pla->normal.z*dir);
						edge[numedges+2].v = getvertex(vertex[edge[e].v].x + pla->normal.x*dir,
																 vertex[edge[e].v].y + pla->normal.y*dir,
																 vertex[edge[e].v].z + pla->normal.z*dir);
						edge[numedges+3].v = edge[e].v;

						edge[numedges  ].n = numedges+1;
						edge[numedges+1].n = numedges+2;
						edge[numedges+2].n = numedges+3;
						edge[numedges+3].n = numedges;

						edge[numedges  ].l = numedges+3;
						edge[numedges+1].l = numedges;
						edge[numedges+2].l = numedges+1;
						edge[numedges+3].l = numedges+2;

						edge[numedges  ].loop = numloops;
						edge[numedges+1].loop = numloops;
						edge[numedges+2].loop = numloops;
						edge[numedges+3].loop = numloops;

						calcplanetexturedefault(numplanes);
						calcplanenormal(numplanes);

						edge[e].o = numedges+1;
						plane[searchplane].n = numplanes;

						numplanes++;
						numloops++;
						numedges += 4;
					}

					e = ee;
				} while (e != startedge);
				l = loop[l].n;
			} while (l != startloop);

				//Actually push searchplane forward of backward
			startloop = l = pla->loophead;
			do
			{
				startedge = e = loop[l].edgehead;
				do
				{
					edge[e].v = getvertex(vertex[edge[e].v].x + pla->normal.x*dir,
												 vertex[edge[e].v].y + pla->normal.y*dir,
												 vertex[edge[e].v].z + pla->normal.z*dir);
					e = edge[e].n;
				} while (e != startedge);
				l = loop[l].n;
			} while (l != startloop);

				//SLOW way to fix edge opposites...
			k = plane[searchplane].region;
			for(i=0;i<numedges;i++)
			{
				if (plane[loop[edge[i].loop].plane].region != k) continue;
				if ((edge[i].v == edge[edge[edge[i].o].n].v) && (edge[edge[i].o].v == edge[edge[i].n].v))
					continue;
				for(j=0;j<numedges;j++)
					if ((edge[i].v == edge[edge[j].n].v) && (edge[j].v == edge[edge[i].n].v))
						if (plane[loop[edge[j].loop].plane].region == k)
							{ edge[i].o = j; edge[j].o = i; break; }
			}


				//Attempt at removing side planes with no area
				//Must make buffer of planes because edges can get moved inside
				//the removecollinears function
			k = 0;
			startloop = l = pla->loophead;
			do
			{
				startedge = e = loop[l].edgehead;
				do
				{
					planelist[k++] = loop[edge[edge[e].o].loop].plane;
					e = edge[e].n;
				} while (e != startedge);
				l = loop[l].n;
			} while (l != startloop);

				//Make sure planelist is in incremental order so when
				//Removecollinears deletes a plane, the moved plane was
				//already processed
			for(i=1;i<k;i++)
				for(j=0;j<i;j++)
					if (planelist[i] < planelist[j])
						swaplong(&planelist[i],&planelist[j]);

			for(j=k-1;j>=0;j--) removecollinears(planelist[j]);

				//SLOW way to fix edge opposites...
			k = plane[searchplane].region;
			for(i=0;i<numedges;i++)
			{
				if (plane[loop[edge[i].loop].plane].region != k) continue;
				if ((edge[i].v == edge[edge[edge[i].o].n].v) && (edge[edge[i].o].v == edge[edge[i].n].v))
					continue;
				for(j=0;j<numedges;j++)
					if ((edge[i].v == edge[edge[j].n].v) && (edge[j].v == edge[edge[i].n].v))
						if (plane[loop[edge[j].loop].plane].region == k)
							{ edge[i].o = j; edge[j].o = i; break; }
			}

			regionsort(pla->region);
		}
	}
	uninitkeys();
	uninittimer42();
	uninittimer();

	if (vgacompatible)  //Uninit VR flag
		koutpw(0x3d4,(((long)overtbits+32)<<8)+0x11);

	uninitvesa();
	setvmode(0x3);
}

drawline256 (float x1, float y1, float x2, float y2, long col)
{
	float dx, dy;
	long i, j, inc, iend;

	dx = x2-x1; dy = y2-y1; if ((dx == 0) && (dy == 0)) return;
	if (x1 >= fxdim)  { if (x2 >= fxdim) return; y1 += (fxdim-x1)*dy/dx; x1 = fxdim-.000001; }
	else if (x1 < 0) { if (x2 < 0) return; y1 += (0-x1)*dy/dx; x1 = 0; }
	if (x2 >= fxdim)  { y2 += (fxdim-x2)*dy/dx; x2 = fxdim-.000001; }
	else if (x2 < 0) { y2 += (0-x2)*dy/dx; x2 = 0; }
	if (y1 >= fydim)  { if (y2 >= fydim) return; x1 += (fydim-y1)*dx/dy; y1 = fydim-.000001; }
	else if (y1 < 0) { if (y2 < 0) return; x1 += (0-y1)*dx/dy; y1 = 0; }
	if (y2 >= fydim)  { x2 += (fydim-y2)*dx/dy; y2 = fydim-.000001; }
	else if (y2 < 0) { x2 += (0-y2)*dx/dy; y2 = 0; }

	if (fabs(dx) >= fabs(dy))
	{
		if (x2 > x1) { i = x1; iend = x2; } else { i = x2; iend = x1; }
		inc = 1048576.0*dy/dx; j = y1*1048576.0+((float)i+.5-x1)*inc;
		for(;i<=iend;i++,j+=inc)
			if ((unsigned long)(j>>20) < (unsigned long)ydim)
				*(char *)(frameplace+ylookup[j>>20]+i) = (char)col;
	}
	else
	{
		if (y2 > y1) { i = y1; iend = y2; } else { i = y2; iend = y1; }
		inc = 1048576.0*dx/dy; j = x1*1048576.0+((float)i+.5-y1)*inc;
		for(;i<=iend;i++,j+=inc)
			if ((unsigned long)(j>>20) < (unsigned long)xdim)
				*(char *)(frameplace+ylookup[i]+(j>>20)) = (char)col;
	}
}

drawline3d(point3d *v1, point3d *v2, long col)
{
	float ox, oy, oz, x1, y1, z1, x2, y2, z2, r;

	ox = v1->x - ipos.x;
	oy = v1->y - ipos.y;
	oz = v1->z - ipos.z;
	x1 = (ox*istr.x + oy*istr.y + oz*istr.z)*fhxdim;
	y1 = (ox*ihei.x + oy*ihei.y + oz*ihei.z)*fhxdimyxaspect;
	z1 = ox*ifor.x + oy*ifor.y + oz*ifor.z;

	ox = v2->x - ipos.x;
	oy = v2->y - ipos.y;
	oz = v2->z - ipos.z;
	x2 = (ox*istr.x + oy*istr.y + oz*istr.z)*fhxdim;
	y2 = (ox*ihei.x + oy*ihei.y + oz*ihei.z)*fhxdimyxaspect;
	z2 = ox*ifor.x + oy*ifor.y + oz*ifor.z;

	if (z1 < SCISSORDIST)
	{
		if (z2 < SCISSORDIST) return;
		r = (SCISSORDIST-z1)/(z2-z1); x1 += (x2-x1)*r; y1 += (y2-y1)*r;
		r = 1.0/z2; x2 *= r; y2 *= r;
	}
	else if (z2 < SCISSORDIST)
	{
		r = (z2-SCISSORDIST)/(z2-z1); x2 += (x1-x2)*r; y2 += (y1-y2)*r;
		r = 1.0/z1; x1 *= r; y1 *= r;
	}
	else
	{
		r = 1.0/z1; x1 *= r; y1 *= r;
		r = 1.0/z2; x2 *= r; y2 *= r;
	}

	drawline256(x1+fhxdim,y1+fhydim,x2+fhxdim,y2+fhydim,col);
}

startspan (long tag)
{
	long i;

	spcnt = 0;
	for(i=0;i<MAXREGIONS;i++)
	{
		spbuf[spcnt] = 0x7fffffff;   //Must be highest integer to exit loop
		spbuf[spcnt+1] = 0x7fffffff; //Must be highest integer to exit loop
		spbuf[spcnt+2] = spcnt;
		spcnt += 3;
	}
	if (bytesperline == xdim)
	{
		spbuf[tag*3+2] = spcnt;
		spbuf[spcnt] = 0;
		spbuf[spcnt+1] = ylookup[ydim];
		spbuf[spcnt+2] = tag*3;
		spcnt += 3;
	}
	else
	{
		spbuf[tag*3+2] = spcnt;
		for(i=0;i<ydim;i++)
		{
			spbuf[spcnt] = ylookup[i];
			spbuf[spcnt+1] = ylookup[i+1];
			spbuf[spcnt+2] = spcnt+3;
			spcnt += 3;
		}
		spbuf[spcnt-1] = tag*3;
	}
}

initags (long intag, long outag)
{
	spi0 = spi1 = intag*3; spi2 = spbuf[spi0+2];
	spo0 = spo1 = outag*3; spo2 = spbuf[spo0+2];
}

	//Returns plane and point in hit, or plane = -1 if bad
long hitscan (long daregion, point3d *c, point3d *v, point3d *hit)
{
	point3d th;
	float hitdist, nhitdist, top, bot, x1, y1, x2, y2;
	long i, hitplane, r, nr, rcnt, rend, cnt2d;
	long p, startplane, l, startloop, e, startedge;

	if ((unsigned long)daregion >= (unsigned long)numregions) return(0L);

	hit->x = c->x; hit->y = c->y; hit->z = c->z;
	hitdist = -1; hitplane = -1;
	regionlist[0] = daregion; rcnt = 0; rend = 1;
	while (rcnt < rend)
	{
		r = regionlist[rcnt++];

		p = startplane = region[r].planehead;
		do
		{
				//intx = ipos.x + ifor.x*t;
				//inty = ipos.y + ifor.y*t;
				//intz = ipos.z + ifor.z*t;
				//plane[p].normal.x * (intx-vertex[i].x)+
				//plane[p].normal.y * (inty-vertex[i].y)+
				//plane[p].normal.z * (intz-vertex[i].z) = 0;
			bot = plane[p].normal.x*v->x+
					plane[p].normal.y*v->y+
					plane[p].normal.z*v->z;
			if (bot >= 0) { p = plane[p].n; continue; }
			i = edge[loop[plane[p].loophead].edgehead].v;
			top = plane[p].normal.x*(vertex[i].x-c->x)+
					plane[p].normal.y*(vertex[i].y-c->y)+
					plane[p].normal.z*(vertex[i].z-c->z);
			if (top >= 0) { p = plane[p].n; continue; }
			nhitdist = top/bot;
			if ((nhitdist >= hitdist) && (hitdist >= 0))
				{ p = plane[p].n; continue; }

			th.x = v->x*nhitdist + c->x;
			th.y = v->y*nhitdist + c->y;
			th.z = v->z*nhitdist + c->z;
			cnt2d = 0;
			if ((fabs(plane[p].normal.x) > fabs(plane[p].normal.y)) && (fabs(plane[p].normal.x) > fabs(plane[p].normal.z)))
			{
				l = startloop = plane[p].loophead;
				do
				{
					e = startedge = loop[l].edgehead;
					do
					{
						y1 = vertex[edge[e].v].y-th.y; y2 = vertex[edge[edge[e].n].v].y-th.y;
						if ((y1 < 0) != (y2 < 0))
						{
							x1 = vertex[edge[e].v].z-th.z; x2 = vertex[edge[edge[e].n].v].z-th.z;
							if ((x1 < 0) == (x2 < 0))
								{ if (x1 < 0) cnt2d ^= 1; }
							else
								{ if ((x1*y2 < x2*y1) != (y2 < 0)) cnt2d ^= 1; }
						}
						e = edge[e].n;
					} while (e != startedge);
					l = loop[l].n;
				} while (l != startloop);
			}
			else if (fabs(plane[p].normal.y) > fabs(plane[p].normal.z))
			{
				l = startloop = plane[p].loophead;
				do
				{
					e = startedge = loop[l].edgehead;
					do
					{
						y1 = vertex[edge[e].v].z-th.z; y2 = vertex[edge[edge[e].n].v].z-th.z;
						if ((y1 < 0) != (y2 < 0))
						{
							x1 = vertex[edge[e].v].x-th.x; x2 = vertex[edge[edge[e].n].v].x-th.x;
							if ((x1 < 0) == (x2 < 0))
								{ if (x1 < 0) cnt2d ^= 1; }
							else
								{ if ((x1*y2 < x2*y1) != (y2 < 0)) cnt2d ^= 1; }
						}
						e = edge[e].n;
					} while (e != startedge);
					l = loop[l].n;
				} while (l != startloop);
			}
			else
			{
				l = startloop = plane[p].loophead;
				do
				{
					e = startedge = loop[l].edgehead;
					do
					{
						y1 = vertex[edge[e].v].y-th.y; y2 = vertex[edge[edge[e].n].v].y-th.y;
						if ((y1 < 0) != (y2 < 0))
						{
							x1 = vertex[edge[e].v].x-th.x; x2 = vertex[edge[edge[e].n].v].x-th.x;
							if ((x1 < 0) == (x2 < 0))
								{ if (x1 < 0) cnt2d ^= 1; }
							else
								{ if ((x1*y2 < x2*y1) != (y2 < 0)) cnt2d ^= 1; }
						}
						e = edge[e].n;
					} while (e != startedge);
					l = loop[l].n;
				} while (l != startloop);
			}
			if (!(cnt2d&1)) { p = plane[p].n; continue; } //th not inside plane p

			if (plane[p].o >= 0)
			{
				nr = plane[plane[p].o].region;
				for(i=rend-1;i>=0;i--)
					if (regionlist[i] == nr) break;
				if (i < 0) regionlist[rend++] = nr;
			}
			else
			{
				hit->x = th.x; hit->y = th.y; hit->z = th.z; hitdist = nhitdist;
				hitplane = p;
			}

			p = plane[p].n;
		} while (p != startplane);
	}
	return(hitplane);
}

passinitfunc(long daplane)
{
}

passtrapfunc(long day1, long day2)
{
	long y, p1, p2, lop, hip;

	for(y=day1;y<day2;y++)
	{
		if (lastx[y] >= nextx[y]) continue;
		p1 = lastx[y]+ylookup[y]; p2 = nextx[y]+ylookup[y];

		if (spbuf[spi0+2] == spi0) continue;  //intag is empty!
	
			//Find intag indeces
		if (spbuf[spi1+1] > p1) { spi1 = spi0; spi2 = spbuf[spi0+2]; }
		while (spbuf[spi2+1] <= p1) { spi1 = spi2; spi2 = spbuf[spi2+2]; }

		while (spbuf[spi2] < p2)
		{
				//Delete lop-hip from spi0
			if (p1 > spbuf[spi2])
			{
				if (p2 < spbuf[spi2+1])  //Insert spcnt after spi2
				{
					lop = p1; hip = p2;
					spbuf[spcnt] = spbuf[spi2];
					spbuf[spcnt+1] = p1;
					spbuf[spcnt+2] = spi2;
					spbuf[spi2] = p2;
					spbuf[spi1+2] = spcnt;
					spi1 = spcnt;
					spcnt += 3;
				}
				else
				{                        //Nudge right of spi2 to the left
					lop = p1; hip = spbuf[spi2+1];
					spbuf[spi2+1] = p1;
					spi1 = spi2; spi2 = spbuf[spi2+2];
				}
			}
			else
			{
				if (p2 < spbuf[spi2+1])  //Nudge left of spi2 to the right
				{
					lop = spbuf[spi2]; hip = p2;
					spbuf[spi2] = p2;
				}
				else                      //Delete spi2
				{
					lop = spbuf[spi2]; hip = spbuf[spi2+1];
					spbuf[spi1+2] = spbuf[spi2+2];
					spi2 = spbuf[spi2+2];
				}
			}
	
				//Find outag indeces
			if (spbuf[spo1+1] > lop) { spo1 = spo0; spo2 = spbuf[spo0+2]; }
			while (spbuf[spo2] < hip) { spo1 = spo2; spo2 = spbuf[spo2+2]; }
	
				//Insert lop-hip to spo0
			if (spbuf[spo1+1] == lop)
			{
				if (spbuf[spo2] == hip)    //Combine spo1,spo2
				{
					spbuf[spo1+1] = spbuf[spo2+1];
					spbuf[spo1+2] = spbuf[spo2+2];
					spo2 = spbuf[spo2+2];
				}
				else                      //Nudge spo1
					spbuf[spo1+1] = hip;
			}
			else
			{
				if (spbuf[spo2] == hip)   //Nudge spo2
				{
					spbuf[spo2] = lop;
					spo1 = spo2; spo2 = spbuf[spo2+2];
				}
				else                       //Insert spcnt
				{
					spbuf[spcnt+0] = lop;
					spbuf[spcnt+1] = hip;
					spbuf[spcnt+2] = spo2;
					spbuf[spo1+2] = spcnt;
					spo2 = spcnt;
					spcnt += 3;
				}
			}
	
			//Draw lop < p < hip
		}
	}
}

soliinitfunc(long daplane)
{
	globcol = (((daplane<<2)+96)&255);
	globcol |= (globcol<<8); globcol |= (globcol<<16);
}

solitrapfunc(long day1, long day2)
{
	long y, lop, hip, p1, p2;

	for(y=day1;y<day2;y++)
	{
		if (lastx[y] >= nextx[y]) continue;
		p1 = lastx[y]+ylookup[y]; p2 = nextx[y]+ylookup[y];

		if (spbuf[spi0+2] == spi0) return;  //intag is empty!

			//Find intag indeces
		if (spbuf[spi1+1] > p1) { spi1 = spi0; spi2 = spbuf[spi0+2]; }
		while (spbuf[spi2+1] <= p1) { spi1 = spi2; spi2 = spbuf[spi2+2]; }

		while (spbuf[spi2] < p2)
		{
				//Delete lop-hip from spi0
			if (p1 > spbuf[spi2])
			{
				if (p2 < spbuf[spi2+1])  //Insert spcnt after spi2
				{
					lop = p1; hip = p2;
					spbuf[spcnt] = spbuf[spi2];
					spbuf[spcnt+1] = p1;
					spbuf[spcnt+2] = spi2;
					spbuf[spi2] = p2;
					spbuf[spi1+2] = spcnt;
					spi1 = spcnt;
					spcnt += 3;
				}
				else
				{                        //Nudge right of spi2 to the left
					lop = p1; hip = spbuf[spi2+1];
					spbuf[spi2+1] = p1;
					spi1 = spi2; spi2 = spbuf[spi2+2];
				}
			}
			else
			{
				if (p2 < spbuf[spi2+1])  //Nudge left of spi2 to the right
				{
					lop = spbuf[spi2]; hip = p2;
					spbuf[spi2] = p2;
				}
				else                      //Delete spi2
				{
					lop = spbuf[spi2]; hip = spbuf[spi2+1];
					spbuf[spi1+2] = spbuf[spi2+2];
					spi2 = spbuf[spi2+2];
				}
			}

				//draw lop to hip-1
			clearbufbyte(frameplace+lop,hip-lop,globcol);
			*(char *)(frameplace+lop) = whitecol;
		}
	}
}

textinitfunc(long daplane)
{
	float ox, oy, oz, t, fx, fy;
	planetype *pla;

	pla = (planetype *)&plane[daplane];

	switch(pla->alignmethod)
	{
		case 0:
			rp[2].x -= rp[1].x; rp[2].y -= rp[1].y; rp[2].z -= rp[1].z;
			rp[0].x -= rp[1].x; rp[0].y -= rp[1].y; rp[0].z -= rp[1].z;
			break;
		case 1:
				//normal closest to <0,0,1>
			ox = 1; oy = 0; oz = -pla->normal.x/pla->normal.z;
			rp[2].x = ox*istr.x + oy*istr.y + oz*istr.z;
			rp[2].y = ox*ihei.x + oy*ihei.y + oz*ihei.z;
			rp[2].z = ox*ifor.x + oy*ifor.y + oz*ifor.z;

			ox = 0; oy = 1; oz = -pla->normal.y/pla->normal.z;
			rp[0].x = ox*istr.x + oy*istr.y + oz*istr.z;
			rp[0].y = ox*ihei.x + oy*ihei.y + oz*ihei.z;
			rp[0].z = ox*ifor.x + oy*ifor.y + oz*ifor.z;

			break;
		case 2:
				//normal closest to <0,1,0>
			ox = 1; oy = -pla->normal.x/pla->normal.y; oz = 0;
			rp[2].x = ox*istr.x + oy*istr.y + oz*istr.z;
			rp[2].y = ox*ihei.x + oy*ihei.y + oz*ihei.z;
			rp[2].z = ox*ifor.x + oy*ifor.y + oz*ifor.z;

			ox = 0; oy = -pla->normal.z/pla->normal.y; oz = 1;
			rp[0].x = ox*istr.x + oy*istr.y + oz*istr.z;
			rp[0].y = ox*ihei.x + oy*ihei.y + oz*ihei.z;
			rp[0].z = ox*ifor.x + oy*ifor.y + oz*ifor.z;
			break;
		case 3:
				//normal closest to <1,0,0>
			ox = -pla->normal.y/pla->normal.x; oy = 1; oz = 0;
			rp[2].x = ox*istr.x + oy*istr.y + oz*istr.z;
			rp[2].y = ox*ihei.x + oy*ihei.y + oz*ihei.z;
			rp[2].z = ox*ifor.x + oy*ifor.y + oz*ifor.z;

			ox = -pla->normal.z/pla->normal.x; oy = 0; oz = 1;
			rp[0].x = ox*istr.x + oy*istr.y + oz*istr.z;
			rp[0].y = ox*ihei.x + oy*ihei.y + oz*ihei.z;
			rp[0].z = ox*ifor.x + oy*ifor.y + oz*ifor.z;
			break;
	}

	t = (float)tilesizx[pla->picnum] / ((float)pla->uscale*64.0);
	rp[2].x *= t; rp[2].y *= t; rp[2].z *= t;
	t = (float)tilesizy[pla->picnum] / ((float)pla->vscale*64.0);
	rp[0].x *= t; rp[0].y *= t; rp[0].z *= t;
	fx = ((float)pla->uoffs)/1024.0;
	fy = ((float)pla->voffs)/1024.0;
	rp[1].x -= rp[2].x*fx + rp[0].x*fy;
	rp[1].y -= rp[2].y*fx + rp[0].y*fy;
	rp[1].z -= rp[2].z*fx + rp[0].z*fy;

	gux = (rp[0].z*rp[1].y-rp[0].y*rp[1].z);
	guy = (rp[0].x*rp[1].z-rp[0].z*rp[1].x)*fxyaspect;
	guo = (rp[0].y*rp[1].x-rp[0].x*rp[1].y)*fhxdim;
	gvx = (rp[2].y*rp[1].z-rp[2].z*rp[1].y);
	gvy = (rp[2].z*rp[1].x-rp[2].x*rp[1].z)*fxyaspect;
	gvo = (rp[2].x*rp[1].y-rp[2].y*rp[1].x)*fhxdim;
	gdx = (rp[0].y*rp[2].z-rp[2].y*rp[0].z);
	gdy = (rp[2].x*rp[0].z-rp[0].x*rp[2].z)*fxyaspect;
	gdo = (rp[0].x*rp[2].y-rp[2].x*rp[0].y)*fhxdim;

	fgdxx = gdx*BOXSIZ; fguxx = gux*8; fgvxx = gvx*8;

	incy = incyy = 0x80000000; //Init every time gd, gux, etc. change

	if ((pla->picnum != globpicnum) && (waloff[pla->picnum] != 0) && (picsiz[pla->picnum] != 0))
	{
		globpicnum = pla->picnum;
		setupkspin(picsiz[pla->picnum],waloff[pla->picnum]);
	}
}

texttrapfunc(long day1, long day2)
{
	long x1, x2, y, lop, hip, p1, p2, nincx0, nincx1, nincy, *longptr;

	for(y=day1;y<day2;y++)
	{
		if (lastx[y] >= nextx[y]) continue;
		p1 = lastx[y]+ylookup[y]; p2 = nextx[y]+ylookup[y];


		if (spbuf[spi0+2] == spi0) return;  //intag is empty!

			//Find intag indeces
		if (spbuf[spi1+1] > p1) { spi1 = spi0; spi2 = spbuf[spi0+2]; }
		while (spbuf[spi2+1] <= p1) { spi1 = spi2; spi2 = spbuf[spi2+2]; }

		while (spbuf[spi2] < p2)
		{
				//Delete lop-hip from spi0
			if (p1 > spbuf[spi2])
			{
				if (p2 < spbuf[spi2+1])  //Insert spcnt after spi2
				{
					lop = p1; hip = p2;
					spbuf[spcnt] = spbuf[spi2];
					spbuf[spcnt+1] = p1;
					spbuf[spcnt+2] = spi2;
					spbuf[spi2] = p2;
					spbuf[spi1+2] = spcnt;
					spi1 = spcnt;
					spcnt += 3;
				}
				else
				{                        //Nudge right of spi2 to the left
					lop = p1; hip = spbuf[spi2+1];
					spbuf[spi2+1] = p1;
					spi1 = spi2; spi2 = spbuf[spi2+2];
				}
			}
			else
			{
				if (p2 < spbuf[spi2+1])  //Nudge left of spi2 to the right
				{
					lop = spbuf[spi2]; hip = p2;
					spbuf[spi2] = p2;
				}
				else                      //Delete spi2
				{
					lop = spbuf[spi2]; hip = spbuf[spi2+1];
					spbuf[spi1+2] = spbuf[spi2+2];
					spi2 = spbuf[spi2+2];
				}
			}


				//draw lop to hip-1
			x1 = lop-ylookup[y]; x2 = hip-ylookup[y];

				//Prepare all nearby ubuf and vbuf points
			nincx0 = (x1 & ~(BOXSIZ-1));
			nincx1 = ((x2+(BOXSIZ-1)) & ~(BOXSIZ-1));
			nincy = (y & ~(BOXSIZ-1));
			if (nincy == incy)
			{
				if (nincx0 < incx0)
				{
					getuvbuf(nincx0,incx0-BOXSIZ,nincy);
					getuvbufi(nincx0,incx0-BOXSIZ,nincy+BOXSIZ);
					doratio(y&((BOXSIZ<<1)-1),nincx0,(incx0-nincx0)>>LOGBOXSIZ);
					incx0 = nincx0;
				}
				if (nincx1 > incx1)
				{
					getuvbuf(incx1+BOXSIZ,nincx1,nincy);
					getuvbufi(incx1+BOXSIZ,nincx1,nincy+BOXSIZ);
					doratio(y&((BOXSIZ<<1)-1),incx1+BOXSIZ,(nincx1-incx1)>>LOGBOXSIZ);
					incx1 = nincx1;
				}

				if (y == incyy+1)
					doratioinc(incx0,((incx1-incx0)>>LOGBOXSIZ)+1);
				else if (y != incyy)
					doratio(y&((BOXSIZ<<1)-1),incx0,((incx1-incx0)>>LOGBOXSIZ)+1);
			}
			else
			{
				if (nincy-BOXSIZ == incy)
				{
					if (nincx0 < incx0) getuvbuf(nincx0,incx0-BOXSIZ,nincy);
					if (nincx1 > incx1) getuvbuf(incx1+BOXSIZ,nincx1,nincy);
				}
				else getuvbuf(nincx0,nincx1,nincy);
				getuvbufi(nincx0,nincx1,nincy+BOXSIZ);
				doratio(y&((BOXSIZ<<1)-1),nincx0,((nincx1-nincx0)>>LOGBOXSIZ)+1);
				incx0 = nincx0; incx1 = nincx1; incy = nincy;
			}
			incyy = y;

			longptr = (long *)&ubuf[(x2>>LOGBOXSIZ)<<1];
			if (x2&(BOXSIZ-1))
			{
				kspinshort(min(x2&(BOXSIZ-1),x2-x1),longptr[0]-longptr[2],
					longptr[1]-longptr[3],ylookup[y] + x2-1 + frameplace,
					(longptr[0]<<LOGBOXSIZ)-(longptr[0]-longptr[2])*(x2&(BOXSIZ-1)),
					(longptr[1]<<LOGBOXSIZ)-(longptr[1]-longptr[3])*(x2&(BOXSIZ-1)));

				if ((x2-x1) <= (x2&(BOXSIZ-1))) continue;
			}

			kspin((x2&~(BOXSIZ-1))-((x1+(BOXSIZ-1))&~(BOXSIZ-1)),0L,
				(long)&ubuf[((x2>>LOGBOXSIZ)-1)<<1],
				ylookup[y]+(x2&~(BOXSIZ-1))+frameplace,longptr[0]<<LOGBOXSIZ,
				longptr[1]<<LOGBOXSIZ);

			if (x1&(BOXSIZ-1))
			{
				longptr = (long *)&ubuf[(x1>>LOGBOXSIZ)<<1];
				kspinshort(BOXSIZ-(x1&(BOXSIZ-1)),longptr[0]-longptr[2],
					longptr[1]-longptr[3],ylookup[y] + (x1|(BOXSIZ-1)) + frameplace,
					longptr[2]<<LOGBOXSIZ,longptr[3]<<LOGBOXSIZ);
			}
		}
	}
}

static float const fscissordist = SCISSORDIST;
rasterizeplane (long daplane, void initfunc(long), void trapfunc(long, long))
{
	point3d *rp0, *rp1;
	planetype *pla;
	float ox, oy, oz, r, x3, y3, x4, y4, fx, fy;
	long b, e, i, j, k, l, y, z, zz, zx, gap, startloop, startedge;
	long z1, z2, z3, z4, splitcnt, start2, startnumbunches;
	long startnpoints, npoints, npoints2, day1, day2, dy1, dy2;

	pla = (planetype *)&plane[daplane];

	npoints2 = 0;
	startloop = l = pla->loophead;
	do
	{
		startnpoints = npoints2;
		startedge = e = edge[loop[l].edgehead].l;
		do
		{
			ox = vertex[edge[e].v].x - ipos.x;
			oy = vertex[edge[e].v].y - ipos.y;
			oz = vertex[edge[e].v].z - ipos.z;
			rp[npoints2].x = ox*istr.x + oy*istr.y + oz*istr.z;
			rp[npoints2].y = ox*ihei.x + oy*ihei.y + oz*ihei.z;
			rp[npoints2].z = ox*ifor.x + oy*ifor.y + oz*ifor.z;
			point22[npoints2] = npoints2+1;
			npoints2++;
			e = edge[e].n;
		} while (e != startedge);
		point22[npoints2-1] = startnpoints;
		l = loop[l].n;
	} while (l != startloop);

	npoints = 0; start2 = 0; z = npoints2-1; splitcnt = 0;
	do
	{
		do
		{
			zz = point22[z]; point22[z] = -1;
			rp0 = &rp[z]; rp1 = &rp[zz];
			if (*(long *)&rp0->z >= *(long *)&fscissordist)
			{
				r = fhxdim/rp0->z;
				sx[npoints] = rp0->x*r + fhxdim;
				sy[npoints] = rp0->y*r*fyxaspect + fhydim;
				point2[npoints] = npoints+1; npoints++;
				if (*(long *)&rp1->z < *(long *)&fscissordist)
				{
					r = (SCISSORDIST-rp0->z)/(rp1->z-rp0->z);
					sx[npoints] = (rp0->x + (rp1->x-rp0->x)*r)/SCISSORDIST*fhxdim + fhxdim;
					sy[npoints] = (rp0->y + (rp1->y-rp0->y)*r)/SCISSORDIST*fhxdimyxaspect + fhydim;
					split[splitcnt++] = npoints;
					point2[npoints] = npoints+1; npoints++;
				}
			}
			else if (*(long *)&rp1->z >= *(long *)&fscissordist)
			{
				r = (SCISSORDIST-rp0->z)/(rp1->z-rp0->z);
				sx[npoints] = (rp0->x + (rp1->x-rp0->x)*r)/SCISSORDIST*fhxdim + fhxdim;
				sy[npoints] = (rp0->y + (rp1->y-rp0->y)*r)/SCISSORDIST*fhxdimyxaspect + fhydim;
				point2[npoints] = npoints+1; npoints++;
			}
			z = zz;
		} while (point22[z] >= 0);

		if (npoints >= start2+3)
			point2[npoints-1] = start2, start2 = npoints;
		else
			npoints = start2;

		do { z--; } while ((z >= 2) && (point22[z] < 0));
	} while (z >= 2);
	if (npoints < 3) return;

	for(z=1;z<splitcnt;z++)
	{
		z1 = split[z]; z2 = point2[z1];
		for(zz=0;zz<z;zz++)
		{
			z3 = split[zz]; z4 = point2[z3];
			r = fabs(sx[z1]-sx[z4])+fabs(sy[z1]-sy[z4])
				+fabs(sx[z3]-sx[z2])+fabs(sy[z3]-sy[z2])
				-fabs(sx[z1]-sx[z2])-fabs(sy[z1]-sy[z2])
				-fabs(sx[z3]-sx[z4])-fabs(sy[z3]-sy[z4]);
			if (*(long *)&r < 0) swaplong(&point2[z1],&point2[z3]);
		}
	}

	clearbufbyte(got,npoints,0L);  //Get bunchindex & numbunches
	z = npoints-1; numbunches = 0;
	do
	{
		startnumbunches = numbunches;
		do
		{
			got[z] = 1;
			zz = point2[z];
				//if ((sy[zz] < sy[z]) != (sy[point2[zz]] < sy[zz]))
			fx = sy[zz]-sy[z]; fy = sy[point2[zz]]-sy[zz];
			if (((*(long *)&fx)^(*(long *)&fy)) < 0)  //Warning: x^y<0 is x^(y<0)
			{
				bunchindex[numbunches] = numbunches;
				bunch[numbunches].v = zz;
				bunch[numbunches].l = numbunches-1;
				numbunches++;
			}
			z = zz;
		} while (!got[z]);
		bunch[startnumbunches].l = numbunches-1;
		do { z--; } while ((z >= 2) && (got[z]));
	} while (z >= 2);

	if (numbunches == 2)
	{
		//sortlist[0] = (sy[bunch[1].v] > sy[bunch[0].v]);
		r = sy[bunch[1].v]-sy[bunch[0].v];
		sortlist[0] = ((*(long *)&r) > 0);
		sortlist[1] = sortlist[0]^1;
		sortlistnum = 2;
	}
	else
	{
			//Sort bunches by y's
		for(gap=(numbunches>>1);gap!=0;gap>>=1)
			for(z=0;z<numbunches-gap;z++)
				for(zz=z;zz>=0;zz-=gap)
				{
					r = sy[bunch[bunchindex[zz]].v]-sy[bunch[bunchindex[zz+gap]].v];
					if (*(long *)&r <= 0) break;
					swaplong(&bunchindex[zz],&bunchindex[zz+gap]);
				}

			//Generate sortlist
		sortlistnum = zx = 0;
		while (1)
		{
			b = bunchindex[zx++]; i = bunch[b].v;
			if (sy[point2[i]] < sy[i]) { got[b] = got[bunch[b].l] = 2; continue; }

			fx = sx[bunch[b].v]; fy = sy[bunch[b].v];
			for(i=sortlistnum-1;i>=0;i--)
			{
				z = sortlist[i];
				if (got[z] == 1)
				{
					e = bunch[z].v;
					if (sy[point2[e]] >= sy[e])
						{ while (sy[point2[e]] < fy) e = point2[e]; }
					else
						{ while (sy[point2[e]] > fy) e = point2[e]; }
					x3 = sx[e]-fx; x4 = sx[point2[e]]-fx;
					if ((x3 < 0) != (x4 < 0))
					{
						y4 = sy[point2[e]]-fy; y3 = x3*y4-x4*(sy[e]-fy);
						if ((y3 < 0) != (y4 < 0)) break;
					}
					else if (x3 < 0) break;
				}
				sortlist[i+2] = z;
			}

			if (!(i&1)) sortlist[i+1] = b, sortlist[i+2] = bunch[b].l;
					 else sortlist[i+2] = b, sortlist[i+1] = bunch[b].l;
			sortlistnum += 2; if (sortlistnum == numbunches) break;
		}
	}

	initfunc(daplane);

		//Draw bunches
	for(zx=0;zx<sortlistnum;zx++)
	{
		z = bunch[sortlist[zx]].v; zz = point2[z];
		r = sy[zz]-sy[z];
		if (*(long *)&r < 0)
		{
			do
			{
				ftol(sy[zz],&day1); if (day1 < 0) day1 = 0;
				ftol(sy[z],&day2); if (day2 > ydim) day2 = ydim;
				if (day2 <= day1) goto endbunch1;

				r = (sx[zz]-sx[z]) / (sy[zz]-sy[z]);
				if (*(long *)&sx[zz] < 0)
				{
					if (*(long *)&sx[z] < 0) { clearbuf(&lastx[day1],day2-day1,0L); goto endbunch1; }
					ftol(sy[zz]+(-sx[zz])/r,&y);
					if (y < day1) y = day1;
					if (y > day2) y = day2;
					clearbuf(&lastx[day1],y-day1,0L); day1 = y;
				}
				else if (*(long *)&sx[z] < 0)
				{
					ftol(sy[zz]+(-sx[zz])/r,&y);
					if (y < day1) y = day1;
					if (y > day2) y = day2;
					clearbuf(&lastx[y],day2-y,0L); day2 = y;
				}
				if (*(long *)&sx[zz] > *(long *)&fxdim)
				{
					if (*(long *)&sx[z] > *(long *)&fxdim) { clearbuf(&lastx[day1],day2-day1,xdim); goto endbunch1; }
					ftol(sy[zz]+(fxdim-sx[zz])/r,&y);
					if (y < day1) y = day1;
					if (y > day2) y = day2;
					clearbuf(&lastx[day1],y-day1,xdim); day1 = y;
				}
				else if (*(long *)&sx[z] > *(long *)&fxdim)
				{
					ftol(sy[zz]+(fxdim-sx[zz])/r,&y);
					if (y < day1) y = day1;
					if (y > day2) y = day2;
					clearbuf(&lastx[y],day2-y,xdim); day2 = y;
				}
				if (day2 > day1)
				{
					ftol((sx[zz]+(((float)day1)-sy[zz])*r)*65536.0,&j);
					ftol(r*65536.0,&k);
					qinterpolatedown16(&lastx[day1],day2-day1,j,k);
				}
endbunch1:; z = zz; zz = point2[z]; r = sy[z]-sy[zz];
			} while (*(long *)&r >= 0);
		}
		else
		{
			do
			{
				ftol(sy[z],&day1); if (day1 < 0) day1 = 0;
				ftol(sy[zz],&day2); if (day2 > ydim) day2 = ydim;
				if (day2 <= day1) goto endbunch2;
				dy1 = day1; dy2 = day2;

				r = (sx[zz]-sx[z]) / (sy[zz]-sy[z]);
				if (*(long *)&sx[z] < 0)
				{
					if (*(long *)&sx[zz] < 0) goto endbunch2;
					ftol(sy[z]+(-sx[z])/r,&y);
					dy1 = day1 = min(max(y,day1),day2);
				}
				else if (*(long *)&sx[zz] < 0)
				{
					ftol(sy[z]+(-sx[z])/r,&y);
					dy2 = day2 = min(max(y,day1),day2);
				}
				if (*(long *)&sx[z] > *(long *)&fxdim)
				{
					if (*(long *)&sx[zz] > *(long *)&fxdim) { clearbuf(&nextx[day1],day2-day1,xdim); day2 = day1; }
					ftol(sy[z]+(fxdim-sx[z])/r,&y);
					if (y < day1) y = day1;
					if (y > day2) y = day2;
					clearbuf(&nextx[day1],y-day1,xdim); day1 = y;
				}
				else if (*(long *)&sx[zz] > *(long *)&fxdim)
				{
					ftol(sy[z]+(fxdim-sx[z])/r,&y);
					if (y < day1) y = day1;
					if (y > day2) y = day2;
					clearbuf(&nextx[y],day2-y,xdim); day2 = y;
				}
				if (day2 > day1)
				{
					ftol((sx[z]+(((float)day1)-sy[z])*r)*65536.0,&j);
					ftol(r*65536.0,&k);
					qinterpolatedown16(&nextx[day1],day2-day1,j,k);
				}
				trapfunc(dy1,dy2);
endbunch2:; z = zz; zz = point2[z]; r = sy[z]-sy[zz];
			} while (*(long *)&r <= 0);
		}
	}
}

	//Returns: testplane is: 0-coplanar, 1-front, 2-back, 3-front&back
planefront(long baseplane, long testplane)
{
	point3d *p, *n, *v;
	float t;
	long i, l, startloop, e, startedge;

	p = &vertex[edge[loop[plane[baseplane].loophead].edgehead].v];
	n = &plane[baseplane].normal;

	i = 0;
	l = startloop = plane[testplane].loophead;
	do
	{
		e = startedge = loop[l].edgehead;
		do
		{
			v = &vertex[edge[e].v];
			t = (p->x-v->x)*n->x + (p->y-v->y)*n->y + (p->z-v->z)*n->z;

			if (*(signed long *)&t > 0x38800000) i |= 1;        //(t > +1/16384)
			else if (*(unsigned long *)&t > 0xb8800000) i |= 2; //(t < -1/16384)

			e = edge[e].n;
		} while (e != startedge);
		l = loop[l].n;
	} while (l != startloop);
	return(i);
}

drawbehindplane(long baseplane, long usetag)
{
	long i, r, p, startplane, curtag;

	curtag = gtagcnt++;
	initags(usetag,curtag);
	rasterizeplane(baseplane,passinitfunc,passtrapfunc);
	if ((spbuf[curtag*3+2] == curtag*3) || (oneregionmode)) return;

	r = plane[plane[baseplane].o].region;
	gotregion[r>>3] |= pow2char[r&7];

	p = startplane = region[r].planehead;
	do
	{
		i = edge[loop[plane[p].loophead].edgehead].v;
		if (plane[p].normal.x*(ipos.x-vertex[i].x)+
			 plane[p].normal.y*(ipos.y-vertex[i].y)+
			 plane[p].normal.z*(ipos.z-vertex[i].z) > 0)
		{
			if (plane[p].o >= 0)
			{
				if ((planefront(baseplane,p) == 1) || (planefront(p,baseplane) == 2))
					drawbehindplane(p,curtag);
			}
			else
			{
				if (planefront(baseplane,p)&1)
				{
					initags(curtag,0);
					if (!gtexturemode) rasterizeplane(p,soliinitfunc,solitrapfunc);
									  else rasterizeplane(p,textinitfunc,texttrapfunc);
				}
			}
			if (spbuf[curtag*3+2] == curtag*3) return;
		}
		p = plane[p].n;
	} while (p != startplane);
}

drawscreen()  //SLOW! - Doesn't combine back sectors!
{
	long i, p, startplane;

	if ((unsigned long)curregion >= (unsigned long)numregions) return;

	setuprotate3d();
	clearbufbyte(gotregion,(numregions+7)>>3,0L);
	gotregion[curregion>>3] |= pow2char[curregion&7];
	startspan(0); gtagcnt = 1;

	startplane = region[curregion].planehead;
	p = startplane;
	do
	{
		i = edge[loop[plane[p].loophead].edgehead].v;
		if (plane[p].normal.x*(ipos.x-vertex[i].x)+
			 plane[p].normal.y*(ipos.y-vertex[i].y)+
			 plane[p].normal.z*(ipos.z-vertex[i].z) > 0)
		{
			if (plane[p].o >= 0)
				drawbehindplane(p,0);
			else
			{
				initags(0,-1);
				if (!gtexturemode) rasterizeplane(p,soliinitfunc,solitrapfunc);
								  else rasterizeplane(p,textinitfunc,texttrapfunc);
			}
		}
		p = plane[p].n;
	} while (p != startplane);
}

initinsertcube(float fx, float fy, float fz)
{
	long i, j, v[8];

	if (numregions >= MAXREGIONS-1) { printf("Too many regions\n"); exit(0); }
	if (numplanes >= MAXPLANES-6) { printf("Too many planes\n"); exit(0); }
	if (numloops >= MAXLOOPS-6) { printf("Too many loops\n"); exit(0); }
	if (numedges >= MAXEDGES-24) { printf("Too many edges\n"); exit(0); }

	region[numregions].planehead = numplanes;
	region[numregions].sortmethod = 0;
	region[numregions].sortvalid = 0;
	region[numregions].tag = 0;
	for(i=5;i>=0;i--)
	{
		plane[numplanes+i].loophead = numloops+i;
		j = i+1; if (j >= 6) j = 0;
		plane[numplanes+i].n = numplanes+j;
		plane[numplanes+i].o = -1;
		plane[numplanes+i].region = numregions;
		loop[numloops+i].edgehead = numedges+i*4;
		loop[numloops+i].n = numloops+i;
		loop[numloops+i].plane = numplanes+i;
	}

	for(i=7;i>=0;i--)
		v[i] = getvertex(fx-32+((i&1)<<6),fy-32+((i&2)<<5),fz-32+((i&4)<<4));

	i = numedges;
	edge[i+ 0].v = v[2]; edge[i+ 0].n = i+1;
	edge[i+ 1].v = v[0]; edge[i+ 1].n = i+2;
	edge[i+ 2].v = v[4]; edge[i+ 2].n = i+3;
	edge[i+ 3].v = v[6]; edge[i+ 3].n = i+0;
	edge[i+ 4].v = v[1]; edge[i+ 4].n = i+5;
	edge[i+ 5].v = v[3]; edge[i+ 5].n = i+6;
	edge[i+ 6].v = v[7]; edge[i+ 6].n = i+7;
	edge[i+ 7].v = v[5]; edge[i+ 7].n = i+4;
	edge[i+ 8].v = v[0]; edge[i+ 8].n = i+9;
	edge[i+ 9].v = v[1]; edge[i+ 9].n = i+10;
	edge[i+10].v = v[5]; edge[i+10].n = i+11;
	edge[i+11].v = v[4]; edge[i+11].n = i+8;
	edge[i+12].v = v[3]; edge[i+12].n = i+13;
	edge[i+13].v = v[2]; edge[i+13].n = i+14;
	edge[i+14].v = v[6]; edge[i+14].n = i+15;
	edge[i+15].v = v[7]; edge[i+15].n = i+12;
	edge[i+16].v = v[2]; edge[i+16].n = i+17;
	edge[i+17].v = v[3]; edge[i+17].n = i+18;
	edge[i+18].v = v[1]; edge[i+18].n = i+19;
	edge[i+19].v = v[0]; edge[i+19].n = i+16;
	edge[i+20].v = v[4]; edge[i+20].n = i+21;
	edge[i+21].v = v[5]; edge[i+21].n = i+22;
	edge[i+22].v = v[7]; edge[i+22].n = i+23;
	edge[i+23].v = v[6]; edge[i+23].n = i+20;

	for(j=23;j>=0;j--) edge[i+j].o = -1;

	numregions++;
	numplanes += 6;
	numloops += 6;
	numedges += 24;
}

krand()
{
	randomseed = ((randomseed*21+1)&65535);
	return(randomseed);
}

#define BSIZ 16
#define MAXSECTORS ((BSIZ*BSIZ*BSIZ)>>2)
static short board[6*BSIZ*BSIZ*BSIZ];    //Board layout
static short tempshort[4096]; //4096 = max(BSIZ*BSIZ*BSIZ,MAXTILES)
static short cubenumsectors;
static short cubesect[BSIZ*BSIZ*BSIZ];
static short sectx1[MAXSECTORS], secty1[MAXSECTORS], sectz1[MAXSECTORS];
static short sectx2[MAXSECTORS], secty2[MAXSECTORS], sectz2[MAXSECTORS];
static short opposect[BSIZ*BSIZ];
static long lin0[BSIZ*BSIZ], lin1[BSIZ*BSIZ];
makeboxbsp()
{
	long i, a, x, y, z, x1, y1, z1, x2, y2, z2, bad, xx, yy, zz;

	a = 0;
	for(i=BSIZ*BSIZ*BSIZ-1;i>=0;i--)
	{
		if (board[i] < 0)
			{ cubesect[i] = -2; tempshort[a++] = i; }
		else
			cubesect[i] = -1;
	}

	cubenumsectors = 0;
	while (1)
	{
		i = mulscale16(krand(),a);
		x = (tempshort[i]/(BSIZ*BSIZ));
		y = ((tempshort[i]/BSIZ)%BSIZ);
		z = (tempshort[i]%BSIZ);
		if (cubesect[tempshort[i]] == -2)
		{
			x1 = x; y1 = y; z1 = z; x2 = x; y2 = y; z2 = z; bad = 0;
			do
			{
				if ((bad&1) == 0)
				{
					x2++;
					for(yy=y1;yy<=y2;yy++)
						for(zz=z1;zz<=z2;zz++)
							if (cubesect[x2*BSIZ*BSIZ+yy*BSIZ+zz] != -2) { bad |= 1; x2--; break; }
				}
				if ((bad&2) == 0)
				{
					y2++;
					for(xx=x1;xx<=x2;xx++)
						for(zz=z1;zz<=z2;zz++)
							if (cubesect[xx*BSIZ*BSIZ+y2*BSIZ+zz] != -2) { bad |= 2; y2--; break; }
				}
				if ((bad&4) == 0)
				{
					z2++;
					for(xx=x1;xx<=x2;xx++)
						for(yy=y1;yy<=y2;yy++)
							if (cubesect[xx*BSIZ*BSIZ+yy*BSIZ+z2] != -2) { bad |= 4; z2--; break; }
				}
				if ((bad&8) == 0)
				{
					x1--;
					for(yy=y1;yy<=y2;yy++)
						for(zz=z1;zz<=z2;zz++)
							if (cubesect[x1*BSIZ*BSIZ+yy*BSIZ+zz] != -2) { bad |= 8; x1++; break; }
				}
				if ((bad&16) == 0)
				{
					y1--;
					for(xx=x1;xx<=x2;xx++)
						for(zz=z1;zz<=z2;zz++)
							if (cubesect[xx*BSIZ*BSIZ+y1*BSIZ+zz] != -2) { bad |= 16; y1++; break; }
				}
				if ((bad&32) == 0)
				{
					z1--;
					for(xx=x1;xx<=x2;xx++)
						for(yy=y1;yy<=y2;yy++)
							if (cubesect[xx*BSIZ*BSIZ+yy*BSIZ+z1] != -2) { bad |= 32; z1++; break; }
				}
			} while (bad != 63);

			for(x=x1;x<=x2;x++)
				for(y=y1;y<=y2;y++)
					for(z=z1;z<=z2;z++)
						cubesect[x*BSIZ*BSIZ+y*BSIZ+z] = cubenumsectors;

			sectx1[cubenumsectors] = x1; sectx2[cubenumsectors] = x2;
			secty1[cubenumsectors] = y1; secty2[cubenumsectors] = y2;
			sectz1[cubenumsectors] = z1; sectz2[cubenumsectors] = z2;
			cubenumsectors++;
		}
		else
		{
			a--; if (a <= 0) return;
			tempshort[i] = tempshort[a];
		}
	}
}

addlinregion(long linecnt, long dapic)
{
	long k, e, l;

	plane[numplanes].loophead = numloops;
	plane[numplanes].n = numplanes+1;
	plane[numplanes].o = -1;
	plane[numplanes].region = numregions;
	plane[numplanes].picnum = dapic;
	plane[numplanes].alignmethod = 0;

		//.l's used temporary marker for loop searching
	for(k=linecnt-1;k>=0;k--) edge[numedges+k].l = 0;
	for(k=linecnt-1;k>=0;k--)
	{
		edge[numedges+k].v = lin0[k];
		for(l=linecnt-1;l>=0;l--)
			if ((lin0[l] == lin1[k]) && (edge[numedges+l].l == 0))
			{
				edge[numedges+k].n = numedges+l;
				edge[numedges+l].l = 1;
				break;
			}
	}
	for(k=linecnt-1;k>=0;k--)
	{
		if (edge[numedges+k].l > 1) continue;
		loop[numloops].edgehead = numedges+k;
		loop[numloops].n = numloops+1;
		loop[numloops].plane = numplanes;

		e = numedges+k;
		do
		{
			edge[e].l = 2;
			e = edge[e].n;
		} while (e != numedges+k);

		numloops++;
	}
	loop[numloops-1].n = plane[numplanes].loophead;
	numplanes++;
	numedges += linecnt;
}

float getceilzofslope(long sectnum, long dax, long day)
{
	float dx, dy, t;
	buildwalltype *wal;

	if (!(buildsector[sectnum].ceilingstat&2)) return((float)buildsector[sectnum].ceilingz);
	wal = &buildwall[buildsector[sectnum].wallptr];
	dx = (float)(buildwall[wal->point2].x-wal->x);
	dy = (float)(buildwall[wal->point2].y-wal->y);
	t = dx*dx+dy*dy; if (t == 0) return((float)buildsector[sectnum].ceilingz);
	t = (dx*(day-wal->y)-dy*(dax-wal->x))/(sqrt(t)*256.0);
	return((float)buildsector[sectnum].ceilingz+(float)buildsector[sectnum].ceilingheinum*t);
}

float getflorzofslope(long sectnum, long dax, long day)
{
	float dx, dy, t;
	buildwalltype *wal;

	if (!(buildsector[sectnum].floorstat&2)) return((float)buildsector[sectnum].floorz);
	wal = &buildwall[buildsector[sectnum].wallptr];
	dx = (float)(buildwall[wal->point2].x-wal->x);
	dy = (float)(buildwall[wal->point2].y-wal->y);
	t = dx*dx+dy*dy; if (t == 0) return((float)buildsector[sectnum].floorz);
	t = (dx*(day-wal->y)-dy*(dax-wal->x))/(sqrt(t)*256.0);
	return((float)buildsector[sectnum].floorz+(float)buildsector[sectnum].floorheinum*t);
}

getzsofslope(long sectnum, long dax, long day, float *ceilz, float *florz)
{
	float dx, dy, t;
	buildwalltype *wal, *wal2;
	buildsectortype *sec;

	sec = &buildsector[sectnum];
	*ceilz = (float)sec->ceilingz; *florz = (float)sec->floorz;
	if ((sec->ceilingstat|sec->floorstat)&2)
	{
		wal = &buildwall[sec->wallptr]; wal2 = &buildwall[wal->point2];
		dx = (float)(wal2->x-wal->x);
		dy = (float)(wal2->y-wal->y);
		t = dx*dx+dy*dy; if (t == 0) return;
		t = (dx*(day-wal->y)-dy*(dax-wal->x))/(sqrt(t)*256.0);
		if (sec->ceilingstat&2) *ceilz = (*ceilz)+(float)sec->ceilingheinum*t;
		if (sec->floorstat&2) *florz = (*florz)+(float)sec->floorheinum*t;
	}
}

addquadplane(long v0, long v1, long v2, long v3)
{
	plane[numplanes].loophead = numloops;
	plane[numplanes].n = numplanes+1;
	plane[numplanes].o = -1;
	plane[numplanes].region = numregions;
	loop[numloops].edgehead = numedges;
	loop[numloops].n = numloops;
	loop[numloops].plane = numplanes;
	edge[numedges  ].v = v0; edge[numedges  ].n = numedges+1;
	edge[numedges+1].v = v1; edge[numedges+1].n = numedges+2;
	edge[numedges+2].v = v2; edge[numedges+2].n = numedges+3;
	edge[numedges+3].v = v3; edge[numedges+3].n = numedges  ;
	numedges += 4; numloops++; numplanes++;

	if (numregions >= MAXREGIONS-1) { printf("Too many regions\n"); exit(0); }
	if (numplanes >= MAXPLANES-6) { printf("Too many planes\n"); exit(0); }
	if (numloops >= MAXLOOPS-6) { printf("Too many loops\n"); exit(0); }
	if (numedges >= MAXEDGES-24) { printf("Too many edges\n"); exit(0); }
}

loadboard(char *daboard)
{
	point3d *v0, *v1, p0, p1;
	float f, fx, fy, fz, mci, mcj, mfi, mfj, nci, ncj, nfi, nfj;
	long i, j, k, x, y, z, fil, posx, posy, posz, a1, a2, a3;
	long opposectcnt, linecnt, startwall, endwall;
	long r, p, startplane, l, startloop, e, startedge;
	short buildcursectnum, buildang;
	char dapalette[1024], tempfilename[80], convertedbuildmap = 0, foundfile = 0;

	numregions = numplanes = numloops = numedges = numvertices = 0;

	strcpy(tempfilename,daboard);
	for(i=strlen(daboard)-1;(i>0) && (daboard[i-1] != '\\');i--);
	if (!strchr(&daboard[i],'.')) strcat(tempfilename,".map");
	j = tempfilename[strlen(tempfilename)-1];
	if (((j == 'P') || (j == 'p')) && ((fil = open(tempfilename,O_BINARY|O_RDONLY,S_IREAD)) != -1))
	{
		convertedbuildmap = 1; foundfile = 1;
		read(fil,&i,4); if (i != 7L) { printf("Wrong map version\n"); exit(0); }
		read(fil,&posx,4);
		read(fil,&posy,4);
		read(fil,&posz,4);
		read(fil,&buildang,2);
		read(fil,&buildcursectnum,2);

		read(fil,&buildnumsectors,2);
		read(fil,buildsector,sizeof(buildsectortype)*buildnumsectors);
		read(fil,&buildnumwalls,2);
		read(fil,buildwall,sizeof(buildwalltype)*buildnumwalls);

		//read(fil,&buildnumsprites,2);
		//read(fil,buildsprite,sizeof(buildspritetype)*buildnumsprites);
		close(fil);

			//Make sure your posz is inside sector
		getzsofslope(buildcursectnum,posx,posy,&nci,&ncj);
		posz = min(max(posz,nci+(4<<8)),ncj-(4<<8));

		ipos.x = +((float)posx)/16.0;
		ipos.y = +((float)posy)/16.0;
		ipos.z = +((float)posz)/256.0;
		ang1 = ((float)buildang+512)*PI/1024.0;
		ang2 = 0;
		ang3 = 0;

		for(r=0;r<buildnumsectors;r++)
		{
			region[numregions].planehead = numplanes;
			region[numregions].sortmethod = 0;
			region[numregions].sortvalid = 0;
			region[numregions].tag = 0;

			startwall = buildsector[r].wallptr;
			endwall = startwall + buildsector[r].wallnum;

				//Go through wall planes...
			for(i=startwall;i<endwall;i++)
			{
				j = buildwall[i].point2;
				getzsofslope(r,buildwall[i].x,buildwall[i].y,&mci,&mfi);
				getzsofslope(r,buildwall[j].x,buildwall[j].y,&mcj,&mfj);
				if (buildwall[i].nextsector >= 0)
				{
					getzsofslope(buildwall[i].nextsector,buildwall[i].x,buildwall[i].y,&nci,&nfi);
					getzsofslope(buildwall[i].nextsector,buildwall[j].x,buildwall[j].y,&ncj,&nfj);
				}
				else
				{
					nci = mci; nfi = mfi;
					ncj = mcj; nfj = mfj;
				}

				if ((nci > mci) || (ncj > mcj))  //Top wall (if necessary)
				{
					plane[numplanes].picnum = buildwall[i].picnum;
					plane[numplanes].alignmethod = 0|0x8000;
					if (tilesizx[buildwall[i].picnum])
					{
						plane[numplanes].uscale = (float)buildwall[i].xrepeat*8.0;
						if (buildwall[i].cstat&8) plane[numplanes].uscale = -plane[numplanes].uscale;
						plane[numplanes].uoffs = buildwall[i].xpanning*65536.0/tilesizx[buildwall[i].picnum];
					}
					if (tilesizy[buildwall[i].picnum])
					{
						plane[numplanes].vscale = (float)buildwall[i].yrepeat*(float)(nci-mci)/2048.0;

						f = buildsector[r].ceilingz-mci;
						if (!(buildwall[i].cstat&4)) // "O" - align to other ceiling
							f += buildsector[buildwall[i].nextsector].ceilingz-buildsector[r].ceilingz;
						plane[numplanes].voffs = buildwall[i].ypanning*256.0-f*32.0*((float)buildwall[i].yrepeat)/((float)tilesizy[buildwall[i].picnum]);

						if (buildwall[i].cstat&256)
						{
							plane[numplanes].vscale = -plane[numplanes].vscale;
							plane[numplanes].voffs = -plane[numplanes].voffs;
						}
					}
					addquadplane(getvertex((float)buildwall[i].x/16.0,(float)buildwall[i].y/16.0,mci/256.0),
									 getvertex((float)buildwall[j].x/16.0,(float)buildwall[j].y/16.0,mcj/256.0),
									 getvertex((float)buildwall[j].x/16.0,(float)buildwall[j].y/16.0,ncj/256.0),
									 getvertex((float)buildwall[i].x/16.0,(float)buildwall[i].y/16.0,nci/256.0));
				}
				if (1)  //Main wall
				{
					plane[numplanes].picnum = buildwall[i].picnum;
					plane[numplanes].alignmethod = 0|0x8000;
					if (tilesizx[buildwall[i].picnum])
					{
						plane[numplanes].uscale = (float)buildwall[i].xrepeat*8.0;
						if (buildwall[i].cstat&8) plane[numplanes].uscale = -plane[numplanes].uscale;
						plane[numplanes].uoffs = buildwall[i].xpanning*65536.0/tilesizx[buildwall[i].picnum];
					}
					if (tilesizy[buildwall[i].picnum])
					{
						plane[numplanes].vscale = (float)buildwall[i].yrepeat*(min(mfi,nfi)-max(mci,nci))/2048.0;

						f = (float)buildsector[r].ceilingz-max(mci,nci);
						if (buildwall[i].cstat&4) // "O" - align to bottom
							f += (float)(buildsector[r].floorz-buildsector[r].ceilingz);
						plane[numplanes].voffs = buildwall[i].ypanning*256.0-f*32.0*((float)buildwall[i].yrepeat)/((float)tilesizy[buildwall[i].picnum]);

						if (buildwall[i].cstat&256)
						{
							plane[numplanes].vscale = -plane[numplanes].vscale;
							plane[numplanes].voffs = -plane[numplanes].voffs;
						}
					}
					addquadplane(getvertex((float)buildwall[i].x/16.0,(float)buildwall[i].y/16.0,max(mci,nci)/256.0),
									 getvertex((float)buildwall[j].x/16.0,(float)buildwall[j].y/16.0,max(mcj,ncj)/256.0),
									 getvertex((float)buildwall[j].x/16.0,(float)buildwall[j].y/16.0,min(mfj,nfj)/256.0),
									 getvertex((float)buildwall[i].x/16.0,(float)buildwall[i].y/16.0,min(mfi,nfi)/256.0));
				}
				if ((nfi < mfi) || (nfj < mfj))  //Bottom wall (if necessary)
				{
					k = i;  //K is index of "2" wall or i if not "2"
					if (buildwall[i].cstat&2)  // "2"
						k = buildwall[i].nextwall;
					if (k < 0) k = i;

					if (tilesizx[buildwall[k].picnum])
					{
						plane[numplanes].uscale = (float)buildwall[i].xrepeat*8.0;
						if (buildwall[i].cstat&8) plane[numplanes].uscale = -plane[numplanes].uscale;
						plane[numplanes].uoffs = buildwall[k].xpanning*65536.0/tilesizx[buildwall[k].picnum];
					}

					plane[numplanes].picnum = buildwall[k].picnum;
					plane[numplanes].alignmethod = 0|0x8000;
					if (tilesizy[buildwall[k].picnum])
					{
						plane[numplanes].vscale = (float)buildwall[k].yrepeat*(float)(mfi-nfi)/2048.0;

						f = (float)buildsector[r].ceilingz-nfi;
						if (!(buildwall[k].cstat&4)) // "O" - align to other floor
							f += (float)(buildsector[buildwall[i].nextsector].floorz-buildsector[r].ceilingz);
						plane[numplanes].voffs = buildwall[k].ypanning*256.0-f*32.0*((float)buildwall[k].yrepeat)/((float)tilesizy[buildwall[k].picnum]);

						if (buildwall[k].cstat&256)
						{
							plane[numplanes].vscale = -plane[numplanes].vscale;
							plane[numplanes].voffs = -plane[numplanes].voffs;
						}
					}

					addquadplane(getvertex((float)buildwall[i].x/16.0,(float)buildwall[i].y/16.0,nfi/256.0),
									 getvertex((float)buildwall[j].x/16.0,(float)buildwall[j].y/16.0,nfj/256.0),
									 getvertex((float)buildwall[j].x/16.0,(float)buildwall[j].y/16.0,mfj/256.0),
									 getvertex((float)buildwall[i].x/16.0,(float)buildwall[i].y/16.0,mfi/256.0));
				}
			}

				//Floor plane
			plane[numplanes].loophead = numloops;
			plane[numplanes].n = numplanes+1;
			plane[numplanes].o = -1;
			plane[numplanes].region = numregions;
			plane[numplanes].picnum = buildsector[r].floorpicnum;
			plane[numplanes].uoffs = 0;
			plane[numplanes].voffs = 0;
			plane[numplanes].uscale = 1;
			plane[numplanes].vscale = 1;
			if (buildsector[r].floorstat&8)
			{
				plane[numplanes].uscale = 2;
				plane[numplanes].vscale = 2;
			}
			if (buildsector[r].floorstat&64)
				plane[numplanes].alignmethod = 0;
			else
				plane[numplanes].alignmethod = 1;
			for(i=startwall;i<endwall;i++)
			{
				mfi = getflorzofslope(r,buildwall[i].x,buildwall[i].y);
				edge[numedges+i-startwall].v = getvertex((float)buildwall[i].x/16.0,(float)buildwall[i].y/16.0,mfi/256.0);
				edge[numedges+i-startwall].n = numedges+buildwall[i].point2-startwall;
				if (buildwall[i].point2 < i)
				{
					loop[numloops].edgehead = numedges+i-startwall;
					loop[numloops].n = numloops+1;
					loop[numloops].plane = numplanes;
					numloops++;
				}
			}
			numedges += endwall-startwall;
			loop[numloops-1].n = plane[numplanes].loophead;
			numplanes++;

				//Ceiling plane
			plane[numplanes].loophead = numloops;
			plane[numplanes].n = numplanes+1;
			plane[numplanes].o = -1;
			plane[numplanes].region = numregions;
			plane[numplanes].picnum = buildsector[r].ceilingpicnum;
			plane[numplanes].uoffs = 0;
			plane[numplanes].voffs = 0;
			plane[numplanes].uscale = 1;
			plane[numplanes].vscale = 1;
			if (buildsector[r].ceilingstat&8)
			{
				plane[numplanes].uscale = 2;
				plane[numplanes].vscale = 2;
			}
			if (buildsector[r].ceilingstat&64)
				plane[numplanes].alignmethod = 0;
			else
				plane[numplanes].alignmethod = 1;
			for(i=startwall;i<endwall;i++)
			{
				mci = getceilzofslope(r,buildwall[i].x,buildwall[i].y);
				edge[numedges+i-startwall].v = getvertex((float)buildwall[i].x/16.0,(float)buildwall[i].y/16.0,mci/256.0);
				edge[numedges+i-startwall].l = numedges+buildwall[i].point2-startwall;
				if (buildwall[i].point2 < i)
				{
					loop[numloops].edgehead = numedges+i-startwall;
					loop[numloops].n = numloops+1;
					loop[numloops].plane = numplanes;
					numloops++;
				}
			}
				//Reverse edge direction for ceiling
			for(i=numedges;i<numedges+endwall-startwall;i++)
				edge[edge[i].l].n = i;
			numedges += endwall-startwall;
			loop[numloops-1].n = plane[numplanes].loophead;
			numplanes++;

			plane[numplanes-1].n = region[numregions].planehead;
			numregions++;

			if (numregions >= MAXREGIONS-1) { printf("Too many regions\n"); exit(0); }
			if (numplanes >= MAXPLANES-6) { printf("Too many planes\n"); exit(0); }
			if (numloops >= MAXLOOPS-6) { printf("Too many loops\n"); exit(0); }
			if (numedges >= MAXEDGES-24) { printf("Too many edges\n"); exit(0); }
		}
	}

	if (foundfile == 0)
	{
		strcpy(tempfilename,daboard);
		for(i=strlen(daboard)-1;(i>0) && (daboard[i-1] != '\\');i--);
		if (!strchr(&daboard[i],'.')) strcat(tempfilename,".cub");
		if ((fil = open(tempfilename,O_BINARY|O_RDONLY,S_IREAD)) != -1)
		{
			foundfile = 1;
			read(fil,&i,4); if (i != 0xcbe) { printf("Wrong cube version\n"); exit(0); }
			read(fil,&posx,4);
			read(fil,&posy,4);
			read(fil,&posz,4);
			read(fil,&a1,4);
			read(fil,&a2,4);
			read(fil,&a3,4);
			read(fil,board,sizeof(board));
			close(fil);

			ipos.x = ((float)+posx)/16-8*64-32;
			ipos.y = -(((float)posz)/16-8*64-32);
			ipos.z = ((float)+posy)/16-8*64-32;
			ang1 = a3*PI/1024.0;
			ang2 = a2*PI/1024.0;
			ang3 = a1*PI/1024.0;
	
			makeboxbsp();
	
			for(i=0;i<BSIZ*BSIZ*BSIZ;i++)
				if (board[i] == -1)
				{
					board[i+BSIZ*BSIZ*BSIZ*1] = -1;
					board[i+BSIZ*BSIZ*BSIZ*2] = -1;
					board[i+BSIZ*BSIZ*BSIZ*3] = -1;
					board[i+BSIZ*BSIZ*BSIZ*4] = -1;
					board[i+BSIZ*BSIZ*BSIZ*5] = -1;
				}
			for(i=cubenumsectors-1;i>=0;i--)
			{
				region[numregions].planehead = numplanes;
				region[numregions].sortmethod = 0;
				region[numregions].sortvalid = 0;
				region[numregions].tag = 0;
	
					//Left (x = -32)
				x = sectx1[i]; opposect[0] = -1; opposectcnt = 1;
				for(y=secty1[i];y<=secty2[i];y++)
					for(z=sectz1[i];z<=sectz2[i];z++)
					{
						for(j=opposectcnt-1;j>=0;j--)
							if (opposect[j] == cubesect[((x-1)<<8)+(y<<4)+z]) break;
						if (j < 0) opposect[opposectcnt++] = cubesect[((x-1)<<8)+(y<<4)+z];
					}
				for(j=opposectcnt-1;j>0;j--)
				{
					linecnt = 0;
					for(y=secty1[i];y<=secty2[i];y++)
						for(z=sectz1[i];z<=sectz2[i];z++)
						{
							if (cubesect[((x-1)<<8)+(y<<4)+z] != opposect[j]) continue;
							fx = (float)(x-8)*64; fy = (float)-(z-8)*64; fz = (float)(y-8)*64;
							if ((y == secty1[i]) || (cubesect[((x-1)<<8)+((y-1)<<4)+z] != opposect[j])) { lin0[linecnt] = getvertex(fx-32,fy+32,fz-32); lin1[linecnt] = getvertex(fx-32,fy-32,fz-32); linecnt++; }
							if ((y == secty2[i]) || (cubesect[((x-1)<<8)+((y+1)<<4)+z] != opposect[j])) { lin0[linecnt] = getvertex(fx-32,fy-32,fz+32); lin1[linecnt] = getvertex(fx-32,fy+32,fz+32); linecnt++; }
							if ((z == sectz1[i]) || (cubesect[((x-1)<<8)+(y<<4)+(z-1)] != opposect[j])) { lin0[linecnt] = getvertex(fx-32,fy+32,fz+32); lin1[linecnt] = getvertex(fx-32,fy+32,fz-32); linecnt++; }
							if ((z == sectz2[i]) || (cubesect[((x-1)<<8)+(y<<4)+(z+1)] != opposect[j])) { lin0[linecnt] = getvertex(fx-32,fy-32,fz-32); lin1[linecnt] = getvertex(fx-32,fy-32,fz+32); linecnt++; }
						}
					addlinregion(linecnt,0L);
				}
				opposect[0] = -1; opposectcnt = 1;
				for(y=secty1[i];y<=secty2[i];y++)
					for(z=sectz1[i];z<=sectz2[i];z++)
					{
						for(j=opposectcnt-1;j>=0;j--)
							if (opposect[j] == board[((x-1)<<8)+(y<<4)+z+(3<<12)]) break;
						if (j < 0) opposect[opposectcnt++] = board[((x-1)<<8)+(y<<4)+z+(3<<12)];
					}
				for(j=opposectcnt-1;j>0;j--)
				{
					linecnt = 0;
					for(y=secty1[i];y<=secty2[i];y++)
						for(z=sectz1[i];z<=sectz2[i];z++)
						{
							if (board[((x-1)<<8)+(y<<4)+z+(3<<12)] != opposect[j]) continue;
							fx = (float)(x-8)*64; fy = (float)-(z-8)*64; fz = (float)(y-8)*64;
							if ((y == secty1[i]) || (board[((x-1)<<8)+((y-1)<<4)+z+(3<<12)] != opposect[j])) { lin0[linecnt] = getvertex(fx-32,fy+32,fz-32); lin1[linecnt] = getvertex(fx-32,fy-32,fz-32); linecnt++; }
							if ((y == secty2[i]) || (board[((x-1)<<8)+((y+1)<<4)+z+(3<<12)] != opposect[j])) { lin0[linecnt] = getvertex(fx-32,fy-32,fz+32); lin1[linecnt] = getvertex(fx-32,fy+32,fz+32); linecnt++; }
							if ((z == sectz1[i]) || (board[((x-1)<<8)+(y<<4)+(z-1)+(3<<12)] != opposect[j])) { lin0[linecnt] = getvertex(fx-32,fy+32,fz+32); lin1[linecnt] = getvertex(fx-32,fy+32,fz-32); linecnt++; }
							if ((z == sectz2[i]) || (board[((x-1)<<8)+(y<<4)+(z+1)+(3<<12)] != opposect[j])) { lin0[linecnt] = getvertex(fx-32,fy-32,fz-32); lin1[linecnt] = getvertex(fx-32,fy-32,fz+32); linecnt++; }
						}
					addlinregion(linecnt,opposect[j]);
				}
	
				  //Right (x = 32)
				x = sectx2[i]; opposect[0] = -1; opposectcnt = 1;
				for(y=secty1[i];y<=secty2[i];y++)
					for(z=sectz1[i];z<=sectz2[i];z++)
					{
						for(j=opposectcnt-1;j>=0;j--)
							if (opposect[j] == cubesect[((x+1)<<8)+(y<<4)+z]) break;
						if (j < 0) opposect[opposectcnt++] = cubesect[((x+1)<<8)+(y<<4)+z];
					}
				for(j=opposectcnt-1;j>0;j--)
				{
					linecnt = 0;
					for(y=secty1[i];y<=secty2[i];y++)
						for(z=sectz1[i];z<=sectz2[i];z++)
						{
							if (cubesect[((x+1)<<8)+(y<<4)+z] != opposect[j]) continue;
							fx = (float)(x-8)*64; fy = (float)-(z-8)*64; fz = (float)(y-8)*64;
							if ((y == secty1[i]) || (cubesect[((x+1)<<8)+((y-1)<<4)+z] != opposect[j])) { lin0[linecnt] = getvertex(fx+32,fy-32,fz-32); lin1[linecnt] = getvertex(fx+32,fy+32,fz-32); linecnt++; }
							if ((y == secty2[i]) || (cubesect[((x+1)<<8)+((y+1)<<4)+z] != opposect[j])) { lin0[linecnt] = getvertex(fx+32,fy+32,fz+32); lin1[linecnt] = getvertex(fx+32,fy-32,fz+32); linecnt++; }
							if ((z == sectz1[i]) || (cubesect[((x+1)<<8)+(y<<4)+(z-1)] != opposect[j])) { lin0[linecnt] = getvertex(fx+32,fy+32,fz-32); lin1[linecnt] = getvertex(fx+32,fy+32,fz+32); linecnt++; }
							if ((z == sectz2[i]) || (cubesect[((x+1)<<8)+(y<<4)+(z+1)] != opposect[j])) { lin0[linecnt] = getvertex(fx+32,fy-32,fz+32); lin1[linecnt] = getvertex(fx+32,fy-32,fz-32); linecnt++; }
						}
					addlinregion(linecnt,0L);
				}
				opposect[0] = -1; opposectcnt = 1;
				for(y=secty1[i];y<=secty2[i];y++)
					for(z=sectz1[i];z<=sectz2[i];z++)
					{
						for(j=opposectcnt-1;j>=0;j--)
							if (opposect[j] == board[((x+1)<<8)+(y<<4)+z+(0<<12)]) break;
						if (j < 0) opposect[opposectcnt++] = board[((x+1)<<8)+(y<<4)+z+(0<<12)];
					}
				for(j=opposectcnt-1;j>0;j--)
				{
					linecnt = 0;
					for(y=secty1[i];y<=secty2[i];y++)
						for(z=sectz1[i];z<=sectz2[i];z++)
						{
							if (board[((x+1)<<8)+(y<<4)+z+(0<<12)] != opposect[j]) continue;
							fx = (float)(x-8)*64; fy = (float)-(z-8)*64; fz = (float)(y-8)*64;
							if ((y == secty1[i]) || (board[((x+1)<<8)+((y-1)<<4)+z+(0<<12)] != opposect[j])) { lin0[linecnt] = getvertex(fx+32,fy-32,fz-32); lin1[linecnt] = getvertex(fx+32,fy+32,fz-32); linecnt++; }
							if ((y == secty2[i]) || (board[((x+1)<<8)+((y+1)<<4)+z+(0<<12)] != opposect[j])) { lin0[linecnt] = getvertex(fx+32,fy+32,fz+32); lin1[linecnt] = getvertex(fx+32,fy-32,fz+32); linecnt++; }
							if ((z == sectz1[i]) || (board[((x+1)<<8)+(y<<4)+(z-1)+(0<<12)] != opposect[j])) { lin0[linecnt] = getvertex(fx+32,fy+32,fz-32); lin1[linecnt] = getvertex(fx+32,fy+32,fz+32); linecnt++; }
							if ((z == sectz2[i]) || (board[((x+1)<<8)+(y<<4)+(z+1)+(0<<12)] != opposect[j])) { lin0[linecnt] = getvertex(fx+32,fy-32,fz+32); lin1[linecnt] = getvertex(fx+32,fy-32,fz-32); linecnt++; }
						}
					addlinregion(linecnt,opposect[j]);
				}
	
					//Back (y = -32)
				z = sectz2[i]; opposect[0] = -1; opposectcnt = 1;
				for(x=sectx1[i];x<=sectx2[i];x++)
					for(y=secty1[i];y<=secty2[i];y++)
					{
						for(j=opposectcnt-1;j>=0;j--)
							if (opposect[j] == cubesect[(x<<8)+(y<<4)+(z+1)]) break;
						if (j < 0) opposect[opposectcnt++] = cubesect[(x<<8)+(y<<4)+(z+1)];
					}
				for(j=opposectcnt-1;j>0;j--)
				{
					linecnt = 0;
					for(x=sectx1[i];x<=sectx2[i];x++)
						for(y=secty1[i];y<=secty2[i];y++)
						{
							if (cubesect[(x<<8)+(y<<4)+(z+1)] != opposect[j]) continue;
							fx = (float)(x-8)*64; fy = (float)-(z-8)*64; fz = (float)(y-8)*64;
							if ((x == sectx1[i]) || (cubesect[((x-1)<<8)+(y<<4)+(z+1)] != opposect[j])) { lin0[linecnt] = getvertex(fx-32,fy-32,fz+32); lin1[linecnt] = getvertex(fx-32,fy-32,fz-32); linecnt++; }
							if ((x == sectx2[i]) || (cubesect[((x+1)<<8)+(y<<4)+(z+1)] != opposect[j])) { lin0[linecnt] = getvertex(fx+32,fy-32,fz-32); lin1[linecnt] = getvertex(fx+32,fy-32,fz+32); linecnt++; }
							if ((y == secty1[i]) || (cubesect[(x<<8)+((y-1)<<4)+(z+1)] != opposect[j])) { lin0[linecnt] = getvertex(fx-32,fy-32,fz-32); lin1[linecnt] = getvertex(fx+32,fy-32,fz-32); linecnt++; }
							if ((y == secty2[i]) || (cubesect[(x<<8)+((y+1)<<4)+(z+1)] != opposect[j])) { lin0[linecnt] = getvertex(fx+32,fy-32,fz+32); lin1[linecnt] = getvertex(fx-32,fy-32,fz+32); linecnt++; }
						}
					addlinregion(linecnt,0L);
				}
				opposect[0] = -1; opposectcnt = 1;
				for(x=sectx1[i];x<=sectx2[i];x++)
					for(y=secty1[i];y<=secty2[i];y++)
					{
						for(j=opposectcnt-1;j>=0;j--)
							if (opposect[j] == board[(x<<8)+(y<<4)+(z+1)+(2<<12)]) break;
						if (j < 0) opposect[opposectcnt++] = board[(x<<8)+(y<<4)+(z+1)+(2<<12)];
					}
				for(j=opposectcnt-1;j>0;j--)
				{
					linecnt = 0;
					for(x=sectx1[i];x<=sectx2[i];x++)
						for(y=secty1[i];y<=secty2[i];y++)
						{
							if (board[(x<<8)+(y<<4)+(z+1)+(2<<12)] != opposect[j]) continue;
							fx = (float)(x-8)*64; fy = (float)-(z-8)*64; fz = (float)(y-8)*64;
							if ((x == sectx1[i]) || (board[((x-1)<<8)+(y<<4)+(z+1)+(2<<12)] != opposect[j])) { lin0[linecnt] = getvertex(fx-32,fy-32,fz+32); lin1[linecnt] = getvertex(fx-32,fy-32,fz-32); linecnt++; }
							if ((x == sectx2[i]) || (board[((x+1)<<8)+(y<<4)+(z+1)+(2<<12)] != opposect[j])) { lin0[linecnt] = getvertex(fx+32,fy-32,fz-32); lin1[linecnt] = getvertex(fx+32,fy-32,fz+32); linecnt++; }
							if ((y == secty1[i]) || (board[(x<<8)+((y-1)<<4)+(z+1)+(2<<12)] != opposect[j])) { lin0[linecnt] = getvertex(fx-32,fy-32,fz-32); lin1[linecnt] = getvertex(fx+32,fy-32,fz-32); linecnt++; }
							if ((y == secty2[i]) || (board[(x<<8)+((y+1)<<4)+(z+1)+(2<<12)] != opposect[j])) { lin0[linecnt] = getvertex(fx+32,fy-32,fz+32); lin1[linecnt] = getvertex(fx-32,fy-32,fz+32); linecnt++; }
						}
					addlinregion(linecnt,opposect[j]);
				}
	
					//Front (y = 32)
				z = sectz1[i]; opposect[0] = -1; opposectcnt = 1;
				for(x=sectx1[i];x<=sectx2[i];x++)
					for(y=secty1[i];y<=secty2[i];y++)
					{
						for(j=opposectcnt-1;j>=0;j--)
							if (opposect[j] == cubesect[(x<<8)+(y<<4)+(z-1)]) break;
						if (j < 0) opposect[opposectcnt++] = cubesect[(x<<8)+(y<<4)+(z-1)];
					}
				for(j=opposectcnt-1;j>0;j--)
				{
					linecnt = 0;
					for(x=sectx1[i];x<=sectx2[i];x++)
						for(y=secty1[i];y<=secty2[i];y++)
						{
							if (cubesect[(x<<8)+(y<<4)+(z-1)] != opposect[j]) continue;
							fx = (float)(x-8)*64; fy = (float)-(z-8)*64; fz = (float)(y-8)*64;
							if ((x == sectx1[i]) || (cubesect[((x-1)<<8)+(y<<4)+(z-1)] != opposect[j])) { lin0[linecnt] = getvertex(fx-32,fy+32,fz-32); lin1[linecnt] = getvertex(fx-32,fy+32,fz+32); linecnt++; }
							if ((x == sectx2[i]) || (cubesect[((x+1)<<8)+(y<<4)+(z-1)] != opposect[j])) { lin0[linecnt] = getvertex(fx+32,fy+32,fz+32); lin1[linecnt] = getvertex(fx+32,fy+32,fz-32); linecnt++; }
							if ((y == secty1[i]) || (cubesect[(x<<8)+((y-1)<<4)+(z-1)] != opposect[j])) { lin0[linecnt] = getvertex(fx+32,fy+32,fz-32); lin1[linecnt] = getvertex(fx-32,fy+32,fz-32); linecnt++; }
							if ((y == secty2[i]) || (cubesect[(x<<8)+((y+1)<<4)+(z-1)] != opposect[j])) { lin0[linecnt] = getvertex(fx-32,fy+32,fz+32); lin1[linecnt] = getvertex(fx+32,fy+32,fz+32); linecnt++; }
						}
					addlinregion(linecnt,0L);
				}
				opposect[0] = -1; opposectcnt = 1;
				for(x=sectx1[i];x<=sectx2[i];x++)
					for(y=secty1[i];y<=secty2[i];y++)
					{
						for(j=opposectcnt-1;j>=0;j--)
							if (opposect[j] == board[(x<<8)+(y<<4)+(z-1)+(5<<12)]) break;
						if (j < 0) opposect[opposectcnt++] = board[(x<<8)+(y<<4)+(z-1)+(5<<12)];
					}
				for(j=opposectcnt-1;j>0;j--)
				{
					linecnt = 0;
					for(x=sectx1[i];x<=sectx2[i];x++)
						for(y=secty1[i];y<=secty2[i];y++)
						{
							if (board[(x<<8)+(y<<4)+(z-1)+(5<<12)] != opposect[j]) continue;
							fx = (float)(x-8)*64; fy = (float)-(z-8)*64; fz = (float)(y-8)*64;
							if ((x == sectx1[i]) || (board[((x-1)<<8)+(y<<4)+(z-1)+(5<<12)] != opposect[j])) { lin0[linecnt] = getvertex(fx-32,fy+32,fz-32); lin1[linecnt] = getvertex(fx-32,fy+32,fz+32); linecnt++; }
							if ((x == sectx2[i]) || (board[((x+1)<<8)+(y<<4)+(z-1)+(5<<12)] != opposect[j])) { lin0[linecnt] = getvertex(fx+32,fy+32,fz+32); lin1[linecnt] = getvertex(fx+32,fy+32,fz-32); linecnt++; }
							if ((y == secty1[i]) || (board[(x<<8)+((y-1)<<4)+(z-1)+(5<<12)] != opposect[j])) { lin0[linecnt] = getvertex(fx+32,fy+32,fz-32); lin1[linecnt] = getvertex(fx-32,fy+32,fz-32); linecnt++; }
							if ((y == secty2[i]) || (board[(x<<8)+((y+1)<<4)+(z-1)+(5<<12)] != opposect[j])) { lin0[linecnt] = getvertex(fx-32,fy+32,fz+32); lin1[linecnt] = getvertex(fx+32,fy+32,fz+32); linecnt++; }
						}
					addlinregion(linecnt,opposect[j]);
				}
	
					//Top (z = -32)
				y = secty1[i]; opposect[0] = -1; opposectcnt = 1;
				for(x=sectx1[i];x<=sectx2[i];x++)
					for(z=sectz1[i];z<=sectz2[i];z++)
					{
						for(j=opposectcnt-1;j>=0;j--)
							if (opposect[j] == cubesect[(x<<8)+((y-1)<<4)+z]) break;
						if (j < 0) opposect[opposectcnt++] = cubesect[(x<<8)+((y-1)<<4)+z];
					}
				for(j=opposectcnt-1;j>0;j--)
				{
					linecnt = 0;
					for(x=sectx1[i];x<=sectx2[i];x++)
						for(z=sectz1[i];z<=sectz2[i];z++)
						{
							if (cubesect[(x<<8)+((y-1)<<4)+z] != opposect[j]) continue;
							fx = (float)(x-8)*64; fy = (float)-(z-8)*64; fz = (float)(y-8)*64;
							if ((x == sectx1[i]) || (cubesect[((x-1)<<8)+((y-1)<<4)+z] != opposect[j])) { lin0[linecnt] = getvertex(fx-32,fy-32,fz-32); lin1[linecnt] = getvertex(fx-32,fy+32,fz-32); linecnt++; }
							if ((x == sectx2[i]) || (cubesect[((x+1)<<8)+((y-1)<<4)+z] != opposect[j])) { lin0[linecnt] = getvertex(fx+32,fy+32,fz-32); lin1[linecnt] = getvertex(fx+32,fy-32,fz-32); linecnt++; }
							if ((z == sectz1[i]) || (cubesect[(x<<8)+((y-1)<<4)+(z-1)] != opposect[j])) { lin0[linecnt] = getvertex(fx-32,fy+32,fz-32); lin1[linecnt] = getvertex(fx+32,fy+32,fz-32); linecnt++; }
							if ((z == sectz2[i]) || (cubesect[(x<<8)+((y-1)<<4)+(z+1)] != opposect[j])) { lin0[linecnt] = getvertex(fx+32,fy-32,fz-32); lin1[linecnt] = getvertex(fx-32,fy-32,fz-32); linecnt++; }
						}
					addlinregion(linecnt,0L);
				}
				opposect[0] = -1; opposectcnt = 1;
				for(x=sectx1[i];x<=sectx2[i];x++)
					for(z=sectz1[i];z<=sectz2[i];z++)
					{
						for(j=opposectcnt-1;j>=0;j--)
							if (opposect[j] == board[(x<<8)+((y-1)<<4)+z+(4<<12)]) break;
						if (j < 0) opposect[opposectcnt++] = board[(x<<8)+((y-1)<<4)+z+(4<<12)];
					}
				for(j=opposectcnt-1;j>0;j--)
				{
					linecnt = 0;
					for(x=sectx1[i];x<=sectx2[i];x++)
						for(z=sectz1[i];z<=sectz2[i];z++)
						{
							if (board[(x<<8)+((y-1)<<4)+z+(4<<12)] != opposect[j]) continue;
							fx = (float)(x-8)*64; fy = (float)-(z-8)*64; fz = (float)(y-8)*64;
							if ((x == sectx1[i]) || (board[((x-1)<<8)+((y-1)<<4)+z+(4<<12)] != opposect[j])) { lin0[linecnt] = getvertex(fx-32,fy-32,fz-32); lin1[linecnt] = getvertex(fx-32,fy+32,fz-32); linecnt++; }
							if ((x == sectx2[i]) || (board[((x+1)<<8)+((y-1)<<4)+z+(4<<12)] != opposect[j])) { lin0[linecnt] = getvertex(fx+32,fy+32,fz-32); lin1[linecnt] = getvertex(fx+32,fy-32,fz-32); linecnt++; }
							if ((z == sectz1[i]) || (board[(x<<8)+((y-1)<<4)+(z-1)+(4<<12)] != opposect[j])) { lin0[linecnt] = getvertex(fx-32,fy+32,fz-32); lin1[linecnt] = getvertex(fx+32,fy+32,fz-32); linecnt++; }
							if ((z == sectz2[i]) || (board[(x<<8)+((y-1)<<4)+(z+1)+(4<<12)] != opposect[j])) { lin0[linecnt] = getvertex(fx+32,fy-32,fz-32); lin1[linecnt] = getvertex(fx-32,fy-32,fz-32); linecnt++; }
						}
					addlinregion(linecnt,opposect[j]);
				}
	
					//Bottom (z = 32)
				y = secty2[i]; opposect[0] = -1; opposectcnt = 1;
				for(x=sectx1[i];x<=sectx2[i];x++)
					for(z=sectz1[i];z<=sectz2[i];z++)
					{
						for(j=opposectcnt-1;j>=0;j--)
							if (opposect[j] == cubesect[(x<<8)+((y+1)<<4)+z]) break;
						if (j < 0) opposect[opposectcnt++] = cubesect[(x<<8)+((y+1)<<4)+z];
					}
				for(j=opposectcnt-1;j>0;j--)
				{
					linecnt = 0;
					for(x=sectx1[i];x<=sectx2[i];x++)
						for(z=sectz1[i];z<=sectz2[i];z++)
						{
							if (cubesect[(x<<8)+((y+1)<<4)+z] != opposect[j]) continue;
							fx = (float)(x-8)*64; fy = (float)-(z-8)*64; fz = (float)(y-8)*64;
							if ((x == sectx1[i]) || (cubesect[((x-1)<<8)+((y+1)<<4)+z] != opposect[j])) { lin0[linecnt] = getvertex(fx-32,fy+32,fz+32); lin1[linecnt] = getvertex(fx-32,fy-32,fz+32); linecnt++; }
							if ((x == sectx2[i]) || (cubesect[((x+1)<<8)+((y+1)<<4)+z] != opposect[j])) { lin0[linecnt] = getvertex(fx+32,fy-32,fz+32); lin1[linecnt] = getvertex(fx+32,fy+32,fz+32); linecnt++; }
							if ((z == sectz1[i]) || (cubesect[(x<<8)+((y+1)<<4)+(z-1)] != opposect[j])) { lin0[linecnt] = getvertex(fx+32,fy+32,fz+32); lin1[linecnt] = getvertex(fx-32,fy+32,fz+32); linecnt++; }
							if ((z == sectz2[i]) || (cubesect[(x<<8)+((y+1)<<4)+(z+1)] != opposect[j])) { lin0[linecnt] = getvertex(fx-32,fy-32,fz+32); lin1[linecnt] = getvertex(fx+32,fy-32,fz+32); linecnt++; }
						}
					addlinregion(linecnt,0L);
				}
				opposect[0] = -1; opposectcnt = 1;
				for(x=sectx1[i];x<=sectx2[i];x++)
					for(z=sectz1[i];z<=sectz2[i];z++)
					{
						for(j=opposectcnt-1;j>=0;j--)
							if (opposect[j] == board[(x<<8)+((y+1)<<4)+z+(1<<12)]) break;
						if (j < 0) opposect[opposectcnt++] = board[(x<<8)+((y+1)<<4)+z+(1<<12)];
					}
				for(j=opposectcnt-1;j>0;j--)
				{
					linecnt = 0;
					for(x=sectx1[i];x<=sectx2[i];x++)
						for(z=sectz1[i];z<=sectz2[i];z++)
						{
							if (board[(x<<8)+((y+1)<<4)+z+(1<<12)] != opposect[j]) continue;
							fx = (float)(x-8)*64; fy = (float)-(z-8)*64; fz = (float)(y-8)*64;
							if ((x == sectx1[i]) || (board[((x-1)<<8)+((y+1)<<4)+z+(1<<12)] != opposect[j])) { lin0[linecnt] = getvertex(fx-32,fy+32,fz+32); lin1[linecnt] = getvertex(fx-32,fy-32,fz+32); linecnt++; }
							if ((x == sectx2[i]) || (board[((x+1)<<8)+((y+1)<<4)+z+(1<<12)] != opposect[j])) { lin0[linecnt] = getvertex(fx+32,fy-32,fz+32); lin1[linecnt] = getvertex(fx+32,fy+32,fz+32); linecnt++; }
							if ((z == sectz1[i]) || (board[(x<<8)+((y+1)<<4)+(z-1)+(1<<12)] != opposect[j])) { lin0[linecnt] = getvertex(fx+32,fy+32,fz+32); lin1[linecnt] = getvertex(fx-32,fy+32,fz+32); linecnt++; }
							if ((z == sectz2[i]) || (board[(x<<8)+((y+1)<<4)+(z+1)+(1<<12)] != opposect[j])) { lin0[linecnt] = getvertex(fx-32,fy-32,fz+32); lin1[linecnt] = getvertex(fx+32,fy-32,fz+32); linecnt++; }
						}
					addlinregion(linecnt,opposect[j]);
				}
	
				plane[numplanes-1].n = region[numregions].planehead;
				numregions++;
	
				if (numregions >= MAXREGIONS) { printf("Too many regions\n"); exit(0); }
				if (numplanes >= MAXPLANES) { printf("Too many planes\n"); exit(0); }
				if (numloops >= MAXLOOPS) { printf("Too many loops\n"); exit(0); }
				if (numedges >= MAXEDGES) { printf("Too many edges\n"); exit(0); }
			}

			//for(x=0;x<16;x++)
			//   for(y=0;y<16;y++)
			//      for(z=0;z<16;z++)
			//         if (board[(x<<8)+(y<<4)+z]&0x8000)
			//            initinsertcube((float)(x-8)*64,(float)-(z-8)*64,(float)(y-8)*64);
		}
	}
	if (foundfile == 0)  //Default board - lousy cube
	{
		initinsertcube(0,0,0);
		ipos.x = 0; ipos.y = 0; ipos.z = 0; ang1 = 0; ang2 = 0; ang3 = 0;
	}

		//Edge loop parent fixing
	for(i=numloops-1;i>=0;i--)
	{
		e = startedge = loop[i].edgehead;
		do
		{
			edge[e].loop = i;
			e = edge[e].n;
		} while (e != startedge);
	}

	for(i=0;i<numedges;i++) edge[i].o = -1;

		//SLOW way to fix edge opposites...
	for(i=0;i<numedges;i++)
	{
		if (edge[i].o >= 0) continue;
		k = plane[loop[edge[i].loop].plane].region;
		//if ((edge[i].v == edge[edge[edge[i].o].n].v) && (edge[edge[i].o].v == edge[edge[i].n].v))
		//   if (plane[loop[edge[j].loop].plane].region == k)
		//      continue;

		p = startplane = region[k].planehead;
		do
		{
			l = startloop = plane[p].loophead;
			do
			{
				e = startedge = loop[l].edgehead;
				do
				{
					if ((edge[i].v == edge[edge[e].n].v) && (edge[e].v == edge[edge[i].n].v))
					{
						edge[i].o = e; edge[e].o = i;
						goto enddofromloadboard;
					}
					e = edge[e].n;
				} while (e != startedge);
				l = loop[l].n;
			} while (l != startloop);
			p = plane[p].n;
		} while (p != startplane);
enddofromloadboard:;
	}

	for(i=0;i<numedges;i++) edge[edge[i].n].l = i;

	for(i=0;i<numplanes;i++) calcplanenormal(i);

		//Plane opposite checking (Current method won't work 100% of time)
	for(i=1;i<numplanes;i++)
	{
		p0.x = 0; p0.y = 0; p0.z = 0;
		l = startloop = plane[i].loophead;
		do
		{
			e = startedge = loop[l].edgehead;
			do
			{
				p0.x += vertex[edge[e].v].x;
				p0.y += vertex[edge[e].v].y;
				p0.z += vertex[edge[e].v].z;
				e = edge[e].n;
			} while (e != startedge);
			l = loop[l].n;
		} while (l != startloop);

		v0 = &vertex[edge[loop[plane[i].loophead].edgehead].v];
		for(j=0;j<i;j++)
		{
			if (fabs(plane[i].normal.x+plane[j].normal.x) >= 0.0001) continue;
			if (fabs(plane[i].normal.y+plane[j].normal.y) >= 0.0001) continue;
			if (fabs(plane[i].normal.z+plane[j].normal.z) >= 0.0001) continue;

			v1 = &vertex[edge[loop[plane[j].loophead].edgehead].v];

			if (fabs((v1->x-v0->x)*plane[i].normal.x+
						(v1->y-v0->y)*plane[i].normal.y+
						(v1->z-v0->z)*plane[i].normal.z) >= 0.0001) continue;

			p1.x = 0; p1.y = 0; p1.z = 0;
			l = startloop = plane[j].loophead;
			do
			{
				e = startedge = loop[l].edgehead;
				do
				{
					p1.x += vertex[edge[e].v].x;
					p1.y += vertex[edge[e].v].y;
					p1.z += vertex[edge[e].v].z;
					e = edge[e].n;
				} while (e != startedge);
				l = loop[l].n;
			} while (l != startloop);

			if ((p0.x == p1.x) && (p0.y == p1.y) && (p0.z == p1.z))
				{ plane[i].o = j; plane[j].o = i; }
		}
	}

	if (!convertedbuildmap)
	{
		for(i=numplanes-1;i>=0;i--) removecollinears(i);
		for(i=numplanes-1;i>=0;i--) removecollinears(i);
	}

	for(i=numplanes-1;i>=0;i--)
	{
		if (plane[i].alignmethod&0x8000) //Bit 15 says don't touch texture
			plane[i].alignmethod &= ~0x8000;
		else
			calcplanetexturedefault(i);
	}

	for(i=numregions-1;i>=0;i--) regionsort(i);

	for(i=numregions-1;i>=0;i--)
		if (insideregion(&ipos,i))
			{ curregion = i; break; }

	/*for(i=0;i<255;i++)
	{
		dapalette[i*4+0] = i>>(8-actualdacbits);  //Blue
		dapalette[i*4+1] = i>>(8-actualdacbits);  //Green
		dapalette[i*4+2] = i>>(8-actualdacbits);  //Red
		dapalette[i*4+3] = 0;
	}
	dapalette[255*4+0] = (176)>>(8-actualdacbits);
	dapalette[255*4+1] = (0  )>>(8-actualdacbits);
	dapalette[255*4+2] = (176)>>(8-actualdacbits);
	dapalette[255*4+3] = 0;
	VBE_setPalette(0,256,dapalette);

	for(y=0;y<TILESIZ;y++)
		for(x=0;x<TILESIZ;x++)
		{
			pic[(x&7)+(x&~7)*TILESIZ+(y<<3)] = 35*cos(x*PI2/TILESIZ)+35*sin(y*PI2/TILESIZ)+15*sin((x+13)*PI*4/TILESIZ)+15*cos((y-4)*PI*6/TILESIZ)+(rand()&15)-8+128;
			//pic[(x&7)+(x&~7)*TILESIZ+(y<<3)] = min(max((((x+y)<<7) + ((x^y)<<5))/TILESIZ+(rand()&7)-4,0),254);
			//if ((x < (TILESIZ>>2)) && (y < (TILESIZ>>2))) pic[(x&7)+(x&~7)*TILESIZ+(y<<3)] += 16;
			//if ((x < (TILESIZ>>1)) && (y < (TILESIZ>>1))) pic[(x&7)+(x&~7)*TILESIZ+(y<<3)] += 16;
			//if ((x >= TILESIZ-(TILESIZ>>2)) && (y >= TILESIZ-(TILESIZ>>2))) pic[(x&7)+(x&~7)*TILESIZ+(y<<3)] -= 16;
			//if ((x >= TILESIZ-(TILESIZ>>1)) && (y >= TILESIZ-(TILESIZ>>1))) pic[(x&7)+(x&~7)*TILESIZ+(y<<3)] -= 16;
		}

	for(x=0;x<16;x++) if ((1<<x) >= TILESIZ) break;
	for(y=0;y<16;y++) if ((1<<y) >= TILESIZ) break;
	setupkspin(x+(y<<8),(long)pic);*/
}

calcplanetexturedefault (long daplane)
{
	long startedge, e;
	float x0, y0, z0, x1, y1, z1;
	point3d *v0, *v1, *v2;
	planetype *pla;

	pla = (planetype *)&plane[daplane];
	pla->vscale = 1;
	pla->uscale = 1;
	pla->uoffs = 0;
	pla->voffs = 0;
	if (pla->alignmethod == 1) return;

		//First try to align texture with origin at top left for walls
	e = startedge = loop[pla->loophead].edgehead;
	do
	{
		v0 = (point3d *)&vertex[edge[edge[e].l].v];
		v1 = (point3d *)&vertex[edge[e].v];
		v2 = (point3d *)&vertex[edge[edge[e].n].v];

		x0 = v0->x - v1->x; y0 = v0->y - v1->y; z0 = v0->z - v1->z;
		x1 = v2->x - v1->x; y1 = v2->y - v1->y; z1 = v2->z - v1->z;

		if ((pla->normal.z == 0) && (z1 == 0) && (z0 > 0))
			if (fabs(y0*z1-z0*y1)+fabs(z0*x1-x0*z1)+fabs(x0*y1-y0*x1) >= 0.001)
			{
					//Plane is good!
				loop[pla->loophead].edgehead = e;
				pla->vscale = (short)(sqrt(x0*x0 + y0*y0 + z0*z0));
				pla->uscale = (short)(sqrt(x1*x1 + y1*y1 + z1*z1));
				pla->uoffs = 0;
				pla->voffs = 0;
				return;
			}

		if ((pla->normal.x == 0) && (pla->normal.y == 0) && (y1 == 0) && (y0 > 0))
			if (fabs(y0*z1-z0*y1)+fabs(z0*x1-x0*z1)+fabs(x0*y1-y0*x1) >= 0.001)
			{
					//Plane is good!
				loop[pla->loophead].edgehead = e;
				pla->vscale = (short)(sqrt(x0*x0 + y0*y0 + z0*z0));
				pla->uscale = (short)(sqrt(x1*x1 + y1*y1 + z1*z1));
				pla->uoffs = 0;
				pla->voffs = 0;
				return;
			}

		e = edge[e].n;
	} while (e != startedge);


	e = startedge = loop[pla->loophead].edgehead;
	do
	{
		v0 = (point3d *)&vertex[edge[edge[e].l].v];
		v1 = (point3d *)&vertex[edge[e].v];
		v2 = (point3d *)&vertex[edge[edge[e].n].v];

		x0 = v0->x - v1->x; y0 = v0->y - v1->y; z0 = v0->z - v1->z;
		x1 = v2->x - v1->x; y1 = v2->y - v1->y; z1 = v2->z - v1->z;

			//Collinear test
		if (fabs(y0*z1-z0*y1)+fabs(z0*x1-x0*z1)+fabs(x0*y1-y0*x1) >= 0.001)
		{
				//Plane is good!
			loop[pla->loophead].edgehead = e;
			pla->vscale = (short)(sqrt(x0*x0 + y0*y0 + z0*z0));
			pla->uscale = (short)(sqrt(x1*x1 + y1*y1 + z1*z1));
			pla->uoffs = 0;
			pla->voffs = 0;
			return;
		}
		e = edge[e].n;
	} while (e != startedge);
}

	//Same as calcplanetexturedefault, but doesn't try to align textures upright
calcplanetexture (long daplane)
{
	long startedge, e;
	float x0, y0, z0, x1, y1, z1;
	point3d *v0, *v1, *v2;
	planetype *pla;

	pla = (planetype *)&plane[daplane];
	pla->vscale = 1;
	pla->uscale = 1;
	pla->uoffs = 0;
	pla->voffs = 0;
	if (pla->alignmethod == 1) return;

	e = startedge = loop[pla->loophead].edgehead;
	do
	{
		v0 = (point3d *)&vertex[edge[edge[e].l].v];
		v1 = (point3d *)&vertex[edge[e].v];
		v2 = (point3d *)&vertex[edge[edge[e].n].v];

		x0 = v0->x - v1->x; y0 = v0->y - v1->y; z0 = v0->z - v1->z;
		x1 = v2->x - v1->x; y1 = v2->y - v1->y; z1 = v2->z - v1->z;

			//Collinear test
		if (fabs(y0*z1-z0*y1)+fabs(z0*x1-x0*z1)+fabs(x0*y1-y0*x1) >= 0.001)
		{
				//Plane is good!
			loop[pla->loophead].edgehead = e;
			pla->vscale = (short)(sqrt(x0*x0 + y0*y0 + z0*z0));
			pla->uscale = (short)(sqrt(x1*x1 + y1*y1 + z1*z1));
			pla->uoffs = 0;
			pla->voffs = 0;
			return;
		}
		e = edge[e].n;
	} while (e != startedge);
}

calcplanenormal (long daplane)
{
	long startedge, startloop, e, l, outsidee;
	float nx, ny, nz, x0, y0, z0, x1, y1, z1, f;
	point3d *v0, *v1, *v2, minp, maxp;

	outsidee = -1;
	startloop = l = plane[daplane].loophead;
	do
	{
		startedge = e = loop[l].edgehead;
		do
		{
			v1 = (point3d *)&vertex[edge[e].v];
			if (outsidee < 0)
			{
				minp.x = v1->x; minp.y = v1->y; minp.z = v1->z;
				maxp.x = v1->x; maxp.y = v1->y; maxp.z = v1->z;
				outsidee = e;
			}
			else
			{
				if (v1->x < minp.x) { minp.x = v1->x; outsidee = e; }
				if (v1->y < minp.y) { minp.y = v1->y; outsidee = e; }
				if (v1->z < minp.z) { minp.z = v1->z; outsidee = e; }
				if (v1->x > maxp.x) { maxp.x = v1->x; outsidee = e; }
				if (v1->y > maxp.y) { maxp.y = v1->y; outsidee = e; }
				if (v1->z > maxp.z) { maxp.z = v1->z; outsidee = e; }
			}
			e = edge[e].n;
		} while (e != startedge);
		l = loop[l].n;
	} while (l != startloop);


	startedge = e = outsidee;
	do
	{
		v0 = (point3d *)&vertex[edge[edge[e].l].v];
		v1 = (point3d *)&vertex[edge[e].v];
		v2 = (point3d *)&vertex[edge[edge[e].n].v];

		x0 = v0->x - v1->x; y0 = v0->y - v1->y; z0 = v0->z - v1->z;
		x1 = v2->x - v1->x; y1 = v2->y - v1->y; z1 = v2->z - v1->z;

		nx = y0*z1 - z0*y1;
		ny = z0*x1 - x0*z1;
		nz = x0*y1 - y0*x1;
		f = nx*nx + ny*ny + nz*nz;
		if (f != 0)
		{
			f = 1.0/sqrt(f);
			plane[daplane].normal.x = nx*f;
			plane[daplane].normal.y = ny*f;
			plane[daplane].normal.z = nz*f;
			return;
		}
		e = edge[e].n;
	} while (e != outsidee);
	plane[daplane].normal.x = 0;
	plane[daplane].normal.y = 0;
	plane[daplane].normal.z = 0;
}

setuprotate3d()
{
	float c1, c2, c3, s1, s2, s3, c1c3, c1s3, s1c3, s1s3;

	fcossin(ang1,&c1,&s1);
	fcossin(ang2,&c2,&s2);
	fcossin(ang3,&c3,&s3);
	c1c3 = c1*c3; c1s3 = c1*s3; s1c3 = s1*c3; s1s3 = s1*s3;
	istr.x = c1c3 + s1s3*s2; istr.y = s1c3 - c1s3*s2; istr.z = -s3*c2;
	ihei.x = c1s3 - s1c3*s2; ihei.y = s1s3 + c1c3*s2; ihei.z = c3*c2;
	ifor.x = s1*c2;          ifor.y = -c1*c2;         ifor.z = s2;
}

inittimer()
{
	koutp(0x43,0x36); koutp(0x40,TIMERRATE&255); koutp(0x40,TIMERRATE>>8);
	oldtimerhandler = _dos_getvect(0x8);
	_disable(); _dos_setvect(0x8, timerhandler); _enable();
}

uninittimer()
{
	koutp(0x43,0x36); koutp(0x40,0); koutp(0x40,0);           //18.2 times/sec
	_disable(); _dos_setvect(0x8, oldtimerhandler); _enable();
}

void __interrupt __far timerhandler()
{
	totalclock++;
	chainintrclock -= TIMERRATE;
	if (chainintrclock < 0)
	{
		chainintrclock += 65536;
		_chain_intr(oldtimerhandler);
	}
	koutp(0x20,0x20);
}

initkeys()
{
	long i;

	for(i=0;i<256;i++) keystatus[i] = 0;
	oldkeyhandler = _dos_getvect(0x9);
	_disable(); _dos_setvect(0x9, keyhandler); _enable();
}

uninitkeys()
{
	short *ptr;

	_dos_setvect(0x9, oldkeyhandler);
		//Turn off shifts to prevent stucks with quitting
	ptr = (short *)0x417; *ptr &= ~0x030f;
}

void __interrupt __far keyhandler()
{
	oldreadch = readch; readch = kinp(0x60);
	keytemp = kinp(0x61); koutp(0x61,keytemp|128); koutp(0x61,keytemp&127);
	if ((readch|1) == 0xe1)
		extended = 128;
	else
	{
		if (oldreadch != readch)
			keystatus[(readch&127)+extended] = ((readch>>7)^1);
		extended = 0;
	}
	koutp(0x20,0x20);
	if (keystatus[0x46])
	{
		koutp(0x43,0x36); koutp(0x40,0); koutp(0x40,0);
		_dos_setvect(0x9, oldkeyhandler);
		_dos_setvect(0x8, oldtimerhandler);
		setvmode(0x3);
		exit(0);
	}
}

faketimerhandler() { }

printsmall(long xpos, long ypos, long forcol, long baccol, char *bufptr)
{
	long i, x, y;
	char *ptr;

	ptr = (char *)(frameplace+ylookup[ypos]+xpos);
	while (*bufptr)
	{
		i = smallfnt[*bufptr++];
		for(y=5;y>=0;y--)
		{
			if (i&8) ptr[ylookup[y]  ] = forcol;
			if (i&4) ptr[ylookup[y]+1] = forcol;
			if (i&2) ptr[ylookup[y]+2] = forcol;
			if (i&1) ptr[ylookup[y]+3] = forcol;
			i >>= 4;
		}
		if (baccol >= 0)
			for(y=5;y>=0;y--)
			{
				if (!(i&8)) ptr[ylookup[y]  ] = baccol;
				if (!(i&4)) ptr[ylookup[y]+1] = baccol;
				if (!(i&2)) ptr[ylookup[y]+2] = baccol;
				if (!(i&1)) ptr[ylookup[y]+3] = baccol;
				i >>= 4;
			}
		ptr += 4;
	}
}

aligntogrid(point3d *v, long daplane, long gridalign)
{
	point3d ovec, uvec, vvec, *v0, *v1, *v2;
	float t, fx, fy;
	long i, x, y;

	i = loop[plane[daplane].loophead].edgehead;
	v0 = &vertex[edge[edge[i].l].v];
	v1 = &vertex[edge[i].v];
	v2 = &vertex[edge[edge[i].n].v];

	if (plane[daplane].uscale) t = 1.0 / (float)plane[daplane].uscale;
	uvec.x = (v2->x-v1->x)*t;
	uvec.y = (v2->y-v1->y)*t;
	uvec.z = (v2->z-v1->z)*t;
	if (plane[daplane].vscale) t = 1.0 / (float)plane[daplane].vscale;
	vvec.x = (v0->x-v1->x)*t;
	vvec.y = (v0->y-v1->y)*t;
	vvec.z = (v0->z-v1->z)*t;
	fx = ((float)plane[daplane].uoffs)/1024.0;
	fy = ((float)plane[daplane].voffs)/1024.0;
	ovec.x = v1->x - uvec.x*fx - vvec.x*fy;
	ovec.y = v1->y - uvec.y*fx - vvec.y*fy;
	ovec.z = v1->z - uvec.z*fx - vvec.z*fy;

	fx = uvec.x*(v->x-ovec.x) + uvec.y*(v->y-ovec.y) + uvec.z*(v->z-ovec.z);
	fy = vvec.x*(v->x-ovec.x) + vvec.y*(v->y-ovec.y) + vvec.z*(v->z-ovec.z);

	t = (uvec.x*uvec.x + uvec.y*uvec.y + uvec.z*uvec.z); if (t != 0) fx /= t;
	t = (vvec.x*vvec.x + vvec.y*vvec.y + vvec.z*vvec.z); if (t != 0) fy /= t;

	if (gridlock)
	{
		ftol(fx/gridalign-.5,&x); fx = (float)x*gridalign;
		ftol(fy/gridalign-.5,&y); fy = (float)y*gridalign;
	}

	v->x = ovec.x + uvec.x*fx + vvec.x*fy;
	v->y = ovec.y + uvec.y*fx + vvec.y*fy;
	v->z = ovec.z + uvec.z*fx + vvec.z*fy;
}

long getvertex(float x, float y, float z)
{
	long i;

	for(i=numvertices-1;i>=0;i--)
	{
		if (vertex[i].x != x) continue;
		if (vertex[i].y != y) continue;
		if (vertex[i].z != z) continue;
		return(i);
	}
	vertex[numvertices].x = x;
	vertex[numvertices].y = y;
	vertex[numvertices].z = z;
	numvertices++;
	if (numvertices >= MAXVERTICES) { printf("Too many vertices\n"); exit(0); }
	return(numvertices-1);
}

cleanvertexlist()
{
	long i, j;

	clearbufbyte(gotvertex,numvertices,0L);
	for(i=numedges-1;i>=0;i--)
		gotvertex[edge[i].v] = 1;
	i = 0;
	while (i < numvertices)
	{
		if (gotvertex[i]) { i++; continue; }
		numvertices--;
		vertex[i] = vertex[numvertices];
		gotvertex[i] = gotvertex[numvertices];
		for(j=numedges-1;j>=0;j--)
			if (edge[j].v == numvertices) edge[j].v = i;
	}
}

removecollinears(long daplane)
{
	long i, j, k, startedge, e, nextedge, startloop, l;
	float x0, y0, z0, x1, y1, z1;
	point3d *v0, *v1, *v2;

	return;

	l = startloop = plane[daplane].loophead;
	do
	{
		e = startedge = loop[l].edgehead;
		do
		{
			nextedge = edge[e].n;

				//Plane has less than 3 points so delete it!
			if (edge[startedge].l == edge[startedge].n) break;

			if (edge[e].v == edge[edge[e].n].v)
			{
					//Kill edge e, but ignore edge[e].o since it is invalid

					//Re-wire collinear edge out of list
				if (startedge == e) startedge = edge[e].l;
				if (loop[l].edgehead == e) loop[l].edgehead = edge[e].l;
				edge[edge[e].l].n = edge[e].n;
				edge[edge[e].n].l = edge[e].l;

					//Delete edge e here!
				numedges--;
				memcpy(&edge[e],&edge[numedges],sizeof(edgetype));
				for(j=numedges-1;j>=0;j--)
				{
					if (edge[j].n == numedges) edge[j].n = e;
					if (edge[j].l == numedges) edge[j].l = e;
					if (edge[j].o == numedges) edge[j].o = e;
				}
				for(j=numloops-1;j>=0;j--)
					if (loop[j].edgehead == numedges)
						loop[j].edgehead = e;

					//I hate ugly cases like these
					//(If the edge just copied from numedges gets deleted)
				if (nextedge == numedges) nextedge = e;
				if (startedge == numedges) startedge = e;

				e = nextedge; continue;
			}

				//Don't kill collinear points that are there for a reason!
			if (edge[edge[e].o].loop != edge[edge[edge[e].l].o].loop)
				{ e = nextedge; continue; }

			v0 = (point3d *)&vertex[edge[edge[e].l].v];
			v1 = (point3d *)&vertex[edge[e].v];
			v2 = (point3d *)&vertex[edge[edge[e].n].v];
			x0 = v0->x - v1->x; y0 = v0->y - v1->y; z0 = v0->z - v1->z;
			x1 = v2->x - v1->x; y1 = v2->y - v1->y; z1 = v2->z - v1->z;

				//Collinear test
			if (fabs(y0*z1-z0*y1)+fabs(z0*x1-x0*z1)+fabs(x0*y1-y0*x1) < 0.001)
			{
					//Re-wire collinear edge out of list
				if (startedge == e) startedge = edge[e].l;
				if (loop[l].edgehead == e) loop[l].edgehead = edge[e].l;
				for(i=numloops-1;i>=0;i--)
					if (loop[i].edgehead == edge[edge[e].l].o)
						loop[i].edgehead = edge[e].o;
				i = edge[edge[e].l].o; k = e;

				//  edge[edge[e].l].o         edge[e].o
				//.-----------------.-----------------.
				//edge[e].l         e                 edge[e].n

				edge[edge[e].l].o = edge[e].o;
				edge[edge[e].o].o = edge[e].l;

				edge[edge[i].l].n = edge[i].n;
				edge[edge[i].n].l = edge[i].l;
				edge[edge[k].l].n = edge[k].n;
				edge[edge[k].n].l = edge[k].l;

					//Delete edge i here!
				numedges--;
				memcpy(&edge[i],&edge[numedges],sizeof(edgetype));
				for(j=numedges-1;j>=0;j--)
				{
					if (edge[j].n == numedges) edge[j].n = i;
					if (edge[j].l == numedges) edge[j].l = i;
					if (edge[j].o == numedges) edge[j].o = i;
				}
				for(j=numloops-1;j>=0;j--)
					if (loop[j].edgehead == numedges)
						loop[j].edgehead = i;

					//I hate ugly cases like these
					//(If the edge just copied from numedges gets deleted)
				if (nextedge == numedges) nextedge = i;
				if (startedge == numedges) startedge = i;
				if (k == numedges) k = i;

					//Delete edge e here!
				numedges--;
				memcpy(&edge[k],&edge[numedges],sizeof(edgetype));
				for(j=numedges-1;j>=0;j--)
				{
					if (edge[j].n == numedges) edge[j].n = k;
					if (edge[j].l == numedges) edge[j].l = k;
					if (edge[j].o == numedges) edge[j].o = k;
				}
				for(j=numloops-1;j>=0;j--)
					if (loop[j].edgehead == numedges)
						loop[j].edgehead = k;

					//I hate ugly cases like these
					//(If the edge just copied from numedges gets deleted)
				if (nextedge == numedges) nextedge = k;
				if (startedge == numedges) startedge = k;

				nextedge = edge[nextedge].l;
			}

			e = nextedge;
		} while (e != startedge);

			//Plane has less than 3 points so delete it!
			//Warning #1: doesn't support multiple loops
			//Warning #2: doesn't delete loops
		if (edge[startedge].l == edge[startedge].n)
		{
				//Delete the 2 last edges here!
			i = startedge; k = edge[startedge].n;

				//Delete edge i here!
			numedges--;
			memcpy(&edge[i],&edge[numedges],sizeof(edgetype));
			for(j=numedges-1;j>=0;j--)
			{
				if (edge[j].n == numedges) edge[j].n = i;
				if (edge[j].l == numedges) edge[j].l = i;
				if (edge[j].o == numedges) edge[j].o = i;
			}
			for(j=numloops-1;j>=0;j--)
				if (loop[j].edgehead == numedges)
					loop[j].edgehead = i;

				//I hate ugly cases like these
				//(If the edge just copied from numedges gets deleted)
			if (k == numedges) k = i;

				//Delete edge e here!
			numedges--;
			memcpy(&edge[k],&edge[numedges],sizeof(edgetype));
			for(j=numedges-1;j>=0;j--)
			{
				if (edge[j].n == numedges) edge[j].n = k;
				if (edge[j].l == numedges) edge[j].l = k;
				if (edge[j].o == numedges) edge[j].o = k;
			}
			for(j=numloops-1;j>=0;j--)
				if (loop[j].edgehead == numedges)
					loop[j].edgehead = k;


				//Re-wire daplane out of region
			for(i=numregions-1;i>=0;i--)
				if (region[i].planehead == daplane)
					region[i].planehead = plane[daplane].n;
			for(i=numplanes-1;i>=0;i--)
				if (plane[i].n == daplane)
					plane[i].n = plane[plane[i].n].n;

				//Delete daplane
			numplanes--;
			memcpy(&plane[daplane],&plane[numplanes],sizeof(planetype));
			for(i=numloops-1;i>=0;i--)
				if (loop[i].plane == numplanes)
					loop[i].plane = daplane;
			for(i=numplanes-1;i>=0;i--)
			{
				if (plane[i].n == numplanes) plane[i].n = daplane;
				if (plane[i].o == numplanes) plane[i].o = daplane;
			}
			for(i=numregions-1;i>=0;i--)
				if (region[i].planehead == numplanes)
					region[i].planehead = daplane;
			return;
		}

		l = loop[l].n;
	} while (l != startloop);
}

lastplane(long startplane)
{
	long p;

	for(p=startplane;plane[p].n!=startplane;p=plane[p].n);

	return(p);
}

static char frontstate[16] = {0,0,0,0,0,0,1,1,0,2,0,2,0,2,1,3};
regionsort(long daregion)
{
	long z, zz, zx, zzx, planelistcnt, e0, e1, olde, stat, startplane;

	planelistcnt = 0;
	z = startplane = region[daregion].planehead;
	do
	{
		planelist[planelistcnt++] = z;
		z = plane[z].n;
	} while (z != startplane);

	e0 = 0; e1 = planelistcnt;
	do
	{
		zx = e0; olde = e1-e0;
		do
		{
			z = planelist[zx]; stat = 0;
			for(zzx=e0;zzx<e1;zzx++)
			{
				if (zx == zzx) continue;
				zz = planelist[zzx];
				stat |= frontstate[(planefront(z,zz)<<2)+planefront(zz,z)];
				if (stat == 3) break;
			}
			if (stat < 2)
				{ swaplong(&planelist[zx],&planelist[e0]); e0++; }
			else if (stat == 2)
				{ e1--; swaplong(&planelist[zx],&planelist[e1]); }
			zx++;
		} while (zx < e1);
		if (e1-e0 == olde)
		{
			for(z=256;z>0;z--) { koutp(97,kinp(97)|2); koutp(97,kinp(97)&~2); }
			region[daregion].sortvalid = 0;
			return;
		}
	} while (e0 < e1);

	region[daregion].planehead = planelist[0];
	for(z=0;z<planelistcnt-1;z++) plane[planelist[z]].n = planelist[z+1];
	plane[planelist[planelistcnt-1]].n = planelist[0];
	region[daregion].sortvalid = 1;
}

getedgeofvertexonplane(point3d *davert, long daplane)
{
	long l, e, ee, startloop, startedge;
	float x0, y0, z0, x1, y1, z1;

	l = startloop = plane[daplane].loophead;
	do
	{
		e = startedge = loop[l].edgehead;
		do
		{
			ee = edge[e].n;

				//Collinear test for splitting plane mode
			x0 = vertex[edge[ee].v].x-davert->x;
			y0 = vertex[edge[ee].v].y-davert->y;
			z0 = vertex[edge[ee].v].z-davert->z;
			x1 = davert->x-vertex[edge[e].v].x;
			y1 = davert->y-vertex[edge[e].v].y;
			z1 = davert->z-vertex[edge[e].v].z;
			if (fabs(y0*z1-z0*y1)+fabs(z0*x1-x0*z1)+fabs(x0*y1-y0*x1) < 0.001)
				if (((x0 < 0) == (x1 < 0)) && ((y0 < 0) == (y1 < 0)) && ((z0 < 0) == (z1 < 0)))
					return(e);

			e = ee;
		} while (e != startedge);
		l = loop[l].n;
	} while (l != startloop);
	return(-1);
}

static float num1048576 = 1048576.0;
getuvbuf (long x1, long x2, long y)
{
	float fx, fy, d, u, v;
	//long i;

	fx = x1-fhxdim; fy = y-fhydim;
	d = gdo + gdx*fx + gdy*fy;
	u = guo + gux*fx + guy*fy;
	v = gvo + gvx*fx + gvy*fy;
	x1 >>= (LOGBOXSIZ-1); x2 >>= (LOGBOXSIZ-1);

	if (!(y&BOXSIZ)) getubufasm1((long)&d,(long)&u,(long)&v,0L,x1,x2);
					else getubufasm2((long)&d,(long)&u,(long)&v,0L,x1,x2);

	/*fpustartdiv(&num1048576,&d);
	if (y&BOXSIZ) { x1 += 512; x2 += 512; }
	for(i=x1;i<=x2;i+=2)
	{
		fdmultol(&u,&v,&uubuf[i]);
		d += fgdxx; u += fguxx; v += fgvxx;
		fpustartdiv(&num1048576,&d);
	}
	fpusucknum();*/
}

getuvbufi (long x1, long x2, long y)
{
	float fx, fy, d, u, v;
	//long i;

	fx = x1-fhxdim; fy = y-fhydim;
	d = gdo + gdx*fx + gdy*fy;
	u = guo + gux*fx + guy*fy;
	v = gvo + gvx*fx + gvy*fy;
	x1 >>= (LOGBOXSIZ-1); x2 >>= (LOGBOXSIZ-1);

	if (!(y&BOXSIZ)) getubufasm1((long)&d,(long)&u,(long)&v,0L,x1,x2);
					else getubufasm2((long)&d,(long)&u,(long)&v,0L,x1,x2);

	/*fpustartdiv(&num1048576,&d);
	if (!(y&BOXSIZ))
	{
		for(i=x1;i<=x2;i+=2)
		{
			fdmultol(&u,&v,&uubuf[i]);
			d += fgdxx; u += fguxx; v += fgvxx;
			fpustartdiv(&num1048576,&d);
			uibuf[i  ] = uubuf[i  ]-uubuf[i+512];
			uibuf[i+1] = uubuf[i+1]-uubuf[i+513];
		}
	}
	else
	{
		for(i=x1;i<=x2;i+=2)
		{
			fdmultol(&u,&v,&uubuf[i+512]);
			d += fgdxx; u += fguxx; v += fgvxx;
			fpustartdiv(&num1048576,&d);
			uibuf[i  ] = uubuf[i+512]-uubuf[i  ];
			uibuf[i+1] = uubuf[i+513]-uubuf[i+1];
		}
	}
	fpusucknum();*/
}

updateregion(point3d *dapos, long *daregion)
{
	long startplane, p, i, j, oregion[4];

	if (insideregion(dapos,*daregion)) return;

	oregion[0] = oregion[1] = -1; j = 0;
	startplane = p = region[*daregion].planehead;
	do
	{
		if (plane[p].o >= 0)
		{
			i = plane[plane[p].o].region;  //Won't re-call inside... for 4 regions
			if ((i != oregion[0]) && (i != oregion[1]) && (i != oregion[2]) && (i != oregion[3]))
			{
				if (insideregion(dapos,i)) { *daregion = i; return; }
				oregion[(j++)&3] = i;
			}
		}
		p = plane[p].n;
	} while (p != startplane);

	for(i=numregions-1;i>=0;i--)
		if (insideregion(dapos,i))
			{ *daregion = i; return; }
}

long insideregion (point3d *c, long daregion)
{
	point3d *v;
	long i, p, startplane, l, startloop, e, startedge;
	float x1, y1, x2, y2;
	char cnt2d, cnt3d;

	if ((unsigned long)daregion >= (unsigned long)numregions) return(0L);

	cnt3d = 0;
	p = startplane = region[daregion].planehead;
	do
	{
		if (!plane[p].normal.z) { p = plane[p].n; continue; }

		cnt2d = 0;
		l = startloop = plane[p].loophead;
		do
		{
			e = startedge = loop[l].edgehead;
			do
			{
				y1 = vertex[edge[e].v].y-c->y; y2 = vertex[edge[edge[e].n].v].y-c->y;
				if ((y1 < 0) != (y2 < 0))
				{
					x1 = vertex[edge[e].v].x-c->x; x2 = vertex[edge[edge[e].n].v].x-c->x;
					if ((x1 < 0) == (x2 < 0))
						{ if (x1 < 0) cnt2d ^= 1; }
					else
						{ if ((x1*y2 < x2*y1) != (y2 < 0)) cnt2d ^= 1; }
				}
				e = edge[e].n;
			} while (e != startedge);
			l = loop[l].n;
		} while (l != startloop);
		if (cnt2d)
		{
			v = &vertex[edge[loop[plane[p].loophead].edgehead].v];
			if (((c->x-v->x)*plane[p].normal.x +
				  (c->y-v->y)*plane[p].normal.y +
				  (c->z-v->z)*plane[p].normal.z < 0) ^ (plane[p].normal.z < 0))
				cnt3d ^= 1;
		}
		p = plane[p].n;
	} while (p != startplane);

	return((long)cnt3d);
}

loadpalette()
{
	long i, j, fil;
	char dapalette[1024];

	if (paletteloaded == 0)
	{
		if ((fil = open("palette.dat",O_BINARY|O_RDWR,S_IREAD)) != -1)
		{
			read(fil,palette,768);
			read(fil,&numpalookups,2);
			read(fil,palookup,numpalookups<<8);

			if (translucloaded == 0)
				if ((transluc = (char *)malloc(65536L)) != 0)
					translucloaded = 1;

			read(fil,transluc,65536);

			blackcol = 0; j = 0x7fffffff;
			for(i=0;i<256;i++)
				if (palette[i*3]+palette[i*3+1]+palette[i*3+2] < j)
				{
					j = palette[i*3]+palette[i*3+1]+palette[i*3+2];
					blackcol = i;
				}

			whitecol = 0; j = 0x80000000;
			for(i=0;i<256;i++)
				if (palette[i*3]+palette[i*3+1]+palette[i*3+2] > j)
				{
					j = palette[i*3]+palette[i*3+1]+palette[i*3+2];
					whitecol = i;
				}

			close(fil);
		}
		paletteloaded = 1;
	}
	for(i=0;i<256;i++)
	{
		dapalette[i*4+0] = palette[i*3+2]<<(actualdacbits-6);  //Blue
		dapalette[i*4+1] = palette[i*3+1]<<(actualdacbits-6);  //Green
		dapalette[i*4+2] = palette[i*3+0]<<(actualdacbits-6);  //Red
		dapalette[i*4+3] = 0;
	}
	VBE_setPalette(0,256,dapalette);
}

loadpics(char *filename)
{
	long i, j, x, y, fil, tilestart, tileend, templongbuf[4];

	if ((fil = open(filename,O_BINARY|O_RDONLY,S_IREAD)) == -1)
		return(-1);

	read(fil,templongbuf,16); if (templongbuf[0] != 1) return(-1);
	tilestart = templongbuf[2];
	tileend = templongbuf[3];
	read(fil,tempshort,(tileend-tilestart+1)<<1);
	for(i=tileend-tilestart;i>=0;i--) tilesizx[i+tilestart] = tempshort[i];
	read(fil,tempshort,(tileend-tilestart+1)<<1);
	for(i=tileend-tilestart;i>=0;i--) tilesizy[i+tilestart] = tempshort[i];
	read(fil,&picanm[tilestart],(tileend-tilestart+1)<<2);

	j = 0;
	for(i=tilestart;i<=tileend;i++) j += tilesizx[i]*tilesizy[i];

	if ((pic = (char *)malloc(j)) == NULL)
		{ printf("Not enough memory for art\n"); exit(0); }
	read(fil,pic,j);
	close(fil);

	j = (long)pic;
	for(i=tilestart;i<=tileend;i++)
		{ waloff[i] = j; j += tilesizx[i]*tilesizy[i]; }
	for(i=tilestart;i<=tileend;i++)
	{
		for(x=0;x<16;x++) if ((1<<x) >= tilesizx[i]) break;
		for(y=0;y<16;y++) if ((1<<y) >= tilesizy[i]) break;
		picsiz[i] = x+(y<<8);

		if ((!(tilesizy[i]&7)) && (tilesizx[i]*tilesizy[i] <= 16384))
		{
			for(j=tilesizx[i]*tilesizy[i]-1;j>=0;j--)
				tempchar[j] = *(char *)(waloff[i]+j);
			for(y=0;y<tilesizy[i];y++)
				for(x=0;x<tilesizx[i];x++)
					*(char *)(waloff[i]+(x&7)+((x&~7)*tilesizy[i])+(y<<3)) = tempchar[x*tilesizy[i]+y];
		}
	}

	return(0);
}
